from .api_common import TOOAPI_Baseclass


class Swift_TOO_Status(TOOAPI_Baseclass):
    """Simple class to describe the status of a submitted TOO API request

    Attributes
    ----------
    jobnumber : int
        TOO API job number
    username : str
        username for TOO API (default 'anonymous')
    shared_secret : str
        shared secret for TOO API (default 'anonymous')
    status : str
        status of API request
    timestamp : datetime
        time request was submitted
    began : datetime
        time request began processing
    completed : datetime
        time request finished processing
    errors : list
        list of error strings assoicated with the request
    warnings : list
        list of warning strings associated with the request
    too_id : list
        For a TOO request, the TOO ID assigned to a new request
    """

    # Core API definitions
    _parameters = ["username", "jobnumber", "too_id", "fetchresult"]
    _attributes = [
        "status",
        "errors",
        "warnings",
        "timestamp",
        "began",
        "completed",
        "result",
    ]
    _local = ["api_name", "shared_secret"]
    api_name = "Swift_TOO_Status"

    def __init__(self, *args, **kwargs):
        """
        Parameters
        ----------
        username : str
            username for TOO API (default 'anonymous')
        shared_secret : str
            shared secret for TOO API (default 'anonymous')
        jobnumber : int
            TOO API job number
        """
        # Required arguments
        self.jobnumber = None
        self.username = "anonymous"
        # Optional arguments
        self.fetchresult = None  # This is only to be used with QueryJob

        # Returned parameters
        self.status = "Unknown"
        self.timestamp = None
        self.began = None
        self.completed = None
        self.errors = list()
        self.warnings = list()
        self.too_id = None
        # Result of QueryJob
        self.result = None

        # Parse argument keywords
        self._parseargs(*args, **kwargs)

        # If all arguments are passed, then submit
        if self.validate():
            self.submit()

    def __eq__(self, value):
        return value == self.status

    def __bool__(self):
        if self.status == "Accepted":
            return True
        else:
            return False

    def validate(self):
        """Validate API submission before submit

        Returns
        -------
        bool
            Was validation successful?
        """
        if self.username and self.jobnumber and self.shared_secret:
            return True
        else:
            return False

    def error(self, error):
        """Add an error to the list of errors"""
        if error not in self.errors:
            self.errors.append(error)

    def warning(self, warning):
        """Add a warning to the list of warnings"""
        if warning not in self.warnings:
            self.warnings.append(warning)

    def clear(self):
        """Reset status"""
        self.__init__()


# Aliases for better PEP8 compliance, and future API updates
Swift_TOOStatus = Swift_TOO_Status
TOOStatus = Swift_TOOStatus



HAS_ASTROPY = False
try:
    import astropy.units as u
    from astropy.coordinates import SkyCoord
    from astropy.coordinates.angles import Latitude, Longitude

    HAS_ASTROPY = True
except ImportError:
    pass


class TOOAPI_SkyCoord:
    """Mixin to support for using a SkyCoord in place of RA/Dec. Note that
    swift_too only support SkyCoords if astropy itself is installed. astropy is
    not a dependency for swift_too so will not get installed if you don't already
    have it."""

    _skycoord = None
    _radius = None
    _ra = None
    _dec = None

    @property
    def skycoord(self):
        """Allow TOO requesters to give an astropy SkyCoord object instead of
        RA/Dec. Handy if you want to do things like submit 1950 coordinates or
        Galactic Coordinates."""
        # Check if the RA/Dec match the SkyCoord, and if they don't modify the skycoord
        if HAS_ASTROPY:
            return SkyCoord(self.ra, self.dec, unit="deg", frame="fk5")
        else:
            raise ImportError("To use skycoord, astropy needs to be installed.")

    @skycoord.setter
    def skycoord(self, sc):
        """Convert the SkyCoord into RA/Dec (J2000) when set."""
        if HAS_ASTROPY:
            if sc is None:
                self._skycoord = None
            elif type(sc) is SkyCoord:
                self._skycoord = sc
                self.ra = sc.fk5.ra.deg
                self.dec = sc.fk5.dec.deg
            else:
                raise TypeError("Needs to be assigned an Astropy SkyCoord")
        else:
            raise ImportError("To use skycoord, astropy needs to be installed.")

    @property
    def ra(self):
        return self._ra

    @ra.setter
    def ra(self, ra):
        if HAS_ASTROPY and (type(ra) is u.quantity.Quantity or type(ra) is Longitude):
            ra = ra.to(u.deg).value
        self._ra = ra

    @property
    def dec(self):
        return self._dec

    @dec.setter
    def dec(self, dec):
        if HAS_ASTROPY and (type(dec) is u.quantity.Quantity or type(dec) is Latitude):
            dec = dec.to(u.deg).value
        self._dec = dec

    @property
    def radius(self):
        return self._radius

    @radius.setter
    def radius(self, radius):
        if HAS_ASTROPY and type(radius) is u.quantity.Quantity:
            radius = radius.to(u.deg).value
        self._radius = radius

    # poserr - used by TOO request for position error in arcmins
    @property
    def poserr(self):
        return self._poserr

    @poserr.setter
    def poserr(self, poserr):
        if HAS_ASTROPY and type(poserr) is u.quantity.Quantity:
            poserr = poserr.to(u.arcmin).value
        self._poserr = poserr

    # Aliases
    RA = ra
    declination = dec



"""The `swift_too` module provides a Python API that gives you almost
everything you need to plan and submit Target of Opportunity (TOOs) requests to
the Neil Gehrels Swift Observatory (hereafter *Swift*).

The module is split into the following main classes:

1. Swift_TOO

This allows Target of Opportunity requests to be constructed and submitted. This
includes basic validation before submission, and querying of the status of the
request. This is used in the case where you either don't want to have to deal
with manual web forms, or you wish to submit a TOO request to Swift based on an
automated trigger, for example, if your telescope detects a new transient, and
you want Swift to observe it ASAP, you can automate the process of TOO
submission using this class. A debug mode is provided to allow you to test the
process of submitting a TOO request to completion, without actually submitting a
TOO.

2. Swift_ObsQuery

This class allows you to query what observations Swift has already performed. It
allows you to search for observations by target ID (a unique identifier for a
target) or Observation ID (a unique identifier for a single observation). Also
it allows for searches around fixed celestial coordinates, the default inputs
for this are J2000 right ascension (RA), declination (dec) and radius in decimal
degrees, however we also support astropy's `SkyCoord` to allow for other
coordinate systems to be used for searches, e.g. Galactic Coordinates. In
addition you can narrow the search to specific date ranges.

3. Swift_VisQuery

J2000 RA/dec (given as a `SkyCoord` or `ra` and `dec` properties in decimal
degrees). Before submitting a TOO request it's important that you understand if
a target can actually be observed by Swift. Targets are typically not observable
by Swift if they are closer than 45 degrees to the Sun, 21 degrees and to the
Moon. However, targets that are close to +/-69 in declination can exhibit
periodic pole constraints also that can cause visibility issues for ~10 day
periods. High resolution visibility can be requested, which gives object
visiblity at 1 minute timescales, and includes Earth occultation and periods
inside the South Atlantic Anomaly. As these are computationally expensive to
calculate, we limit high resolution visibliity requests to 20 days in length.

4. Swift_PlanQuery

Similar to Swift_ObsQuery but in this case queries the Swift observing plan,
otherwise known as the pre-planned science timeline (PPST). This allows the user
to query what was planned to be observed by Swift, and what will be observed in
the future. Please note that due to Swift's quick planning turn around, the
amount of information on future plans will be limited to the next 1-3 days
typically.

5. QueryJob

In addition results of jobs can be fetched using the QueryJob module. This can
be used to fetch results of already processed jobs, or query the current status
of a job. Typically a job will have 4 status values, `Queued`, `Processing`,
`Accepted` or `Rejected`. `Accepted` in this case means that processing has
completed. `Queued` means that the job has been accepted, but not yet processed.
`Processing` means that the server is working on the results, but has not yet
completed. `Rejected` means that the server rejected the job for any number of
reasons, that are reported by in `Swift_TOOStatus` class.

6. UVOT_mode

UVOT modes are typically given as hex modes. This class allows the user to look
up the meaning of a given hex code, which is typically a table of UVOT filters
associated with that mode, along with any configuration parameters, such as the
size of the field of view, whether the data will be taken in event mode, etc.

7. Swift_TOORequests

This class allows querying of TOO Requests submitted to Swift. By default this
will give detail on the most recent 10 TOO requests submitted to Swift.
Alternatively, TOOs info can be fetched by year, ID, RA/Dec/Radius, or within a
set time period. In addition if the `detail` parameter is set to `True`, then
this retrieves all information about a TOO request, including non-public
information, for TOO requests that you submitted, if you supply your `username`
and `shared_secret`. The class essentially is a container for a number of
`Swift_TOORequest` objects. In the case where targets are scheduled in the
Swift planning calendar, a `Swift_Calendar` object will be attached which lists
the scheduled windows, and shows how much exposure was obtained during those
windows.

8. Swift_Calendar

`Swift_Calendar` class allows for querying calendar entries for a given TOO. The
Calendar shows all planned observations for a TOO, along with an estimate of how
much time was actually observed during the calendar window. Note that this is
different from the Swift Plan, insofar as the Swift Calendar lists requested
observations and can go much farther into the future. However, due to Swift's
oversubscription, and other issues, even if an object is in the Calendar, that
is not guarantee that Swift will observe it on that day, only that it is in the
queue to be observed.

9. Swift_GUANO

`Swift_GUANO` provides API access to data obtained by the *Gamma-Ray Urgent
Archiver for Novel Opportunities* (GUANO). GUANO proactively dumps BAT event
data, which would otherwise be lost, based on external triggers. These triggers
include Fast Radio Bursts, Gamma-Ray Bursts from other missions, Gravitational
Wave triggers and Neutrino triggers. `Swift_GUANO` allows to query these dumps
and gives you metadata about the dump.

10. Swift_Data

`Swift_Data` provides an easy interface to download archival or quick-look data
from the Swift Science Data Centers in the USA or UK.

11. Swift_Resolve

`Swift_Resolve` provides an simple interface to resolve target names into
coordinates, leveraging several name resolvers. This class is also called by
other classes when you pass the `name` parameter instead of giving an `ra`,
`dec` or `skycoord`.

12. Swift_Clock

For a given `utctime`, `mettime` or `swifttime` return a `swiftdatetime` object
or objects, with clock correction applied. Primarily used internally as part of
the `clock_correct` method, which applies clock correction to all times in a
given class.

13. Swift_SAA

Fetches times when Swift is passing through the South Atlantic Anomaly (SAA) for
a given time period.

Note that all module names have an alias that excludes the `Swift_` part, so for
example, `Swift_VisQuery` becomes `VisQuery`.

The Swift TOO API is built around a client/server model, in which API
information is exchanged between the client's machine with the Swift API server
in JSON format. Requests are submitted using a signed JWT (e.g. https://jwt.io)
format, and API users are required to register with Swift to use the API
(https://www.swift.psu.edu/toop) to submit TOO requests. JWT are not encrypted,
but they are signed with a "shared secret" to ensure that the requests are
coming from the user they say they are. Note that for all requests that are not
submitting a TOO, the username can be set to `anonymous`. In fact this is the
default if no username is given.

Queries are constructed using Python classes provided by this module, and
submitted to a queue system in which they are processed in a first come, first
served basis. Typically processing requests takes a 10-20 seconds. Status of
requests can be queried, and errors are reported back.
"""

from .api_resolve import Resolve, Swift_Resolve
from .query_job import QueryJob
from .swift_calendar import Calendar, Swift_Calendar
from .swift_clock import Clock, Swift_Clock
from .swift_data import Data, Swift_Data
from .swift_guano import GUANO, Swift_GUANO
from .swift_obsquery import ObsQuery, Swift_ObsQuery
from .swift_planquery import PlanQuery, Swift_PlanQuery
from .swift_requests import Swift_TOO_Requests, Swift_TOORequests, TOORequests
from .swift_saa import SAA, Swift_SAA
from .swift_toorequest import TOO, Swift_TOO, Swift_TOO_Request, TOORequest
from .swift_uvot import Swift_UVOTMode, UVOT_mode, UVOTMode
from .swift_visquery import Swift_VisQuery, VisQuery
from .version import version as __version__



__docformat__ = "restructedtext en"
import pandas as pd
from .dataquery_base import dataQuery
from ..data import GRB as dGRB
from .. import plotLightCurve as lcPlot


class GRBQuery(dataQuery):
    """A request to query a GRB cataloguee.

    This inherits from the dataQuery class, and TO BE WRITTEN.
    """

    def __init__(self, cat="UK_XRT", isAux=False, silent=True, verbose=False):
        super().__init__(silent=silent, verbose=verbose)
        """Create a GRBQuery instance.

        May add auxcat in here soon.

        Parameters
        ----------

        cat : str, optional
            Which GRB catalogue to base this query on
            (default: 'UK_XRT').

        isAux : bool, optional
            **For internal use only** whether this is being set as an
            auxilliary catalogue

        silent : bool, optional
            Whether to suppress all console output (default: ``True``).

        verbose : bool, optional
            Whether to give verbose output for everything
            (default: False; overridden by ``silent``).

        """
        # As with SXPS.py, define the tables per catalogue, although
        # in this case there is only one table per catalogue. They still
        # have to be a list for compatibility with the base class.

        self._tablesByCat = {
            "UK_XRT": [
                "XRTLiveCat",
            ],
            "SDC_GRB": [
                "SDC_Data_Table",
            ],
            "BAT_GRB": [
                "BATGRBCat",
            ],
        }
        # Also need the default table to select when a catalogue is selected.
        self._defaultTablesByCat = {"UK_XRT": "XRTLiveCat", "SDC_GRB": "SDC_Data_Table", "BAT_GRB": "BATGRBCat"}

        self._defaultPosSourceByCat = {
            "UK_XRT": "best",
            "SDC_GRB": "BAT",
            "BAT_GRB": "BAT",
        }

        self._posSources = {
            "UK_XRT": {
                "best": ["RA", "Decl"],
                "Onboard": ["Onboard_RA", "Onboard_Decl"],
                "SPER": ["SPER_RA", "SPER_Decl"],
                "Standard": ["PSF_RA", "PSF_Decl"],
                "Enhanced": ["Enh_RA", "SPER_Decl"],
            },
            "SDC_GRB": {
                "BAT": ["BAT_RA", "BAT_Dec"],
                "XRT": ["XRT_RA", "XRT_Dec"],
                "UVOT": ["UVOT_RA", "UVOT_Dec"],
            },
            "BAT_GRB": {
                "BAT": ["RA_ground", "DEC_ground"],
            },
        }

        self._auxCat = None
        self._nameCol = None

        self._prodData.update(
            {
                "lightCurves": None,
                "spectra": None,
                "burstAnalyser": None,
                "positions": None,
            }
        )

        # self._lightCurves = None
        # self._spectra = None
        # self._burstAnalyser = None
        # self._positions = None

        self._isAuxCat = isAux
        self.cat = cat

        # Now set the selected catalogue to that supplied in the constructor.
        # As whichCat is actually a property dectorator, this will check that
        # the catalogue is valid and sert the tables and default table.
        # self.table = table

    # ----------------------------------------------------------
    # Some properties unique to this class.
    @property
    def cat(self):
        """Which GRB catalogue is to be queried."""
        return self._dbName

    @cat.setter
    def cat(self, dbName):
        if dbName in self.cats:
            self._dbName = dbName
            self._tables = self._tablesByCat[dbName]
            self.table = self._defaultTablesByCat[dbName]
            self.posSource = self._defaultPosSourceByCat[dbName]
        else:
            raise ValueError(f"{dbName} is not a known database.")

    @property
    def cats(self):
        """List of queryable GRB catalogues."""
        return list(self._tablesByCat.keys())

    # Need to be able to select which position we use for the search.
    @property
    def posSource(self):
        if self.verbose:
            print(f"Using {self._posSource} for the positions")
            print(f"RA column: {self.raCol}, Dec column: {self.decCol}")
        return self._posSource

    @posSource.setter
    def posSource(self, source):
        if source not in self._posSources[self.cat]:
            raise ValueError(f"`{source}` is not a valid source of position data")
        self._posSource = source
        self._raCol = self._posSources[self.cat][source][0]
        self._decCol = self._posSources[self.cat][source][1]

    @property
    def posSources(self):
        """Possible sources of position data"""
        return list(self._posSources[self.cat].keys())

    @property
    def auxCat(self):
        """Access to the auxilliary catalogue."""
        return self._auxCat

    @property
    def lightCurves(self):
        """GRB light curves."""
        return self._prodData["lightCurves"]

    @property
    def spectra(self):
        """GRB spectra."""
        return self._prodData["spectra"]

    @property
    def burstAnalyser(self):
        """GRB burst analyser data."""
        return self._prodData["burstAnalyser"]

    @property
    def ban(self):
        """GRB burst analyser data."""
        return self._prodData["burstAnalyser"]

    @property
    def positions(self):
        """GRB positions."""
        return self._prodData["positions"]

    # ------------------------------------------------------------------
    # Some functions which override the parent functions

    # Metadata needs to set a name column as well, as this will be used
    # to merge the cat and aux cat.

    def getMetadata(self):
        """Retrieve the metadata for this catalogue from the server.

        This function calls the base class ``getMetadata()`` function,
        for this query and any aux cat.

        """
        super().getMetadata()
        if "IsNameCol" in self._metadata:
            self._metadata["IsNameCol"] = pd.to_numeric(self._metadata["IsNameCol"])
            tmp = self._metadata.loc[self._metadata["IsNameCol"] == 1]["ColName"]
            if len(tmp) > 0:
                self._nameCol = tmp.iloc[0]
                if len(tmp) > 1 and not self.silent:
                    print(
                        "WARNING: Metadata contains TWO name columns! This may be a bug; "
                        "please notify swift-help@leicester.ac.uk"
                    )

    def addConeSearch(self, **kwargs):
        super().addConeSearch(**kwargs)
        if self.auxCat is not None:
            if not self.silent:
                print("Applying cone search to auxilliary catalogue.")
                self.auxCat.addConeSearch(**kwargs)

    def unlock(self):
        super().unlock()
        if self.auxCat is not None:
            self.auxCat.unlock()

    def submit(self, merge=False):
        """Submit the query.

        This wraps the base class ``submit()`` function, but also
        ensures that if there is an aux cat, that is also submitted,
        and the results filtered to only things which match both
        catalogues.
        """

        # If we have an aux cat, we have to ensure that both this
        # and that include the name column, so we can merge.
        if self.auxCat is not None:
            self._forceNameCol()
            self.auxCat._forceNameCol()

        super().submit()
        if self.auxCat is not None:
            self.auxCat.submit()
            # Now we have to merge...
            i1 = self.results.set_index(self._nameCol).index
            i2 = self.auxCat.results.set_index(self.auxCat._nameCol).index
            tmp = self.results[i1.isin(i2)]
            tmp2 = self.auxCat.results[i2.isin(i1)]
            self._results = tmp
            self.auxCat._results = tmp2
            self._results = tmp

            if merge:
                self.mergeResults()

    def mergeResults(self):
        """Merge the results from this and auxCat."""

        if self.results is None:
            if not self.silent:
                print("Cannot merge - there are no results")
            return

        tmp = self.results.merge(self.auxCat.results, how="inner", left_on=self._nameCol, right_on=self.auxCat._nameCol)
        self._results = tmp

    def reset(self, keepAux=False, **kwargs):
        """Reset this query.

        This resets the query so it can be redefined or rerun. It calls
        ``dataquery_base.reset()`` with the specified kwargs, but
        optionally keeps the aux catalogue (but resets it).

        Parameters
        ----------

        keepAux : bool, optional
            Whether the retain (but reset) the auxilliary catalogue
            (default ``False``).

        """
        super().reset(**kwargs)
        if self.auxCat:
            if keepAux:
                # Note, we pass keepAux since in principle the aux
                # cat can have its own aux cat.
                self.auxCat.reset(keepAux=keepAux, **kwargs)
            else:
                self.removeAuxCat()

    # ------------------------------------------------------------------
    # Functions for managing the aux cat

    def getAuxCat(self):
        """Report whether an auxilliary catalogue has been defined.

        Returns
        -------

        str or None:
            If there is no auxilliary catalogue, this will return
            ``None``, otherwise it will return the name of the
            catalogue.
        """

        if self.auxCat is None:
            return None
        return self.auxCat.cat

    def removeAuxCat(self):
        """Remove the auxilliary catalogue"""

        self._auxCat = None

    def setAuxCat(self, cat, silent=None, verbose=None):
        """Set an auxilliary catalogue.

        An auxilliary catalogue allows you to prepare a query which
        combines data from multiple GRB catalogues.
        If you have a cone search then any cone search will be applied
        to the main and auxilliary catalogues; otherwise you can specify
        filters for each catalogue, and then the results returned by
        ``submit()`` will be only GRBs found in both catalogues.

        Parameters
        ----------

        cat : str,
            Which GRB catalogue to use as the auxilliary. **Cannot be
            the same as this catalogue in the main query.**

        silent : bool or None, optional
            Whether to suppress all console output. If ``None`` then it
            will be set to the current query object's ``silent`` value
            (default: ``None``).

        verbose : bool or None, optional
            Whether to give verbose output for everything. If ``None``
            then it will be set to the current query object's
            ``verbose`` value(default: ``None``).

        """
        if cat == self.cat:
            raise ValueError("Cannot have an auxilliary catalogue the same as myself")
        if silent is None:
            silent = self.silent
        if verbose is None:
            verbose = self.verbose
        self._auxCat = GRBQuery(cat=cat, isAux=True, silent=silent, verbose=verbose)

    def _forceNameCol(self):
        """Ensure that the name column is set to be retrieved."""

        if (self.colsToGet is not None) and (self._nameCol not in self.colsToGet):
            self._addCol(self._nameCol)

    # ------------------------------------------------------------------
    # Functions for getting data or products

    # This will look a bit like the SXPS query module, we will allow the
    # user to decide if things are indexed by the targetID or name.

    # getLightCurves
    # getSpectra
    # getBurstAnalyser
    # saveBurstAnalyser

    # Add stuff for positions if I add those.

    # First a function which everything will call, which will handle the
    # byTarget/byName thing:

    def _handleArgs(self, byName=False, byID=False):
        """Decide whether to query on targetID or name.

        This is an internal helper function for any function that gets
        products related to a query. It decides whether to send the
        GRB name or targetID column with the request, depending first,
        on whether either was explicitly requested, and if not, based
        on which columns have been retrieved.

        Parameters
        ----------

        byName : bool
            Whether to force use of the source name.

        byID : bool
            Whether to force use of the source ID

        Returns
        -------
        whichCol : str
            The name o the column to index on in the returned query.
        whichArg : str
            The name of the argument to send to the data module (GRBName
            or targetID).

        """
        whichCol = None
        whichArg = None

        if byID:
            whichCol = self._targetCol
            whichArg = "targetID"
            if byName:
                raise RuntimeError("You cannot set byName and byID!")
        elif byName:
            whichCol = self._nameCol
            whichArg = "GRBName"
        else:
            cols = {self._nameCol: "GRBName", self._targetCol: "sourceID"}

            for c, n in cols.items():
                if c in self.results.columns:
                    whichCol = c
                    whichArg = n
                    break
            if whichCol is None:
                raise RuntimeError(f"Cannot get source info, as none of the columns: {cols} are in your results.")

        if whichCol not in self.results.columns:
            raise RuntimeError(f"Cannot get source info, as the column `{whichCol}` is not in your results.")

        return (whichCol, whichArg)

    def getLightCurves(self, subset=None, byName=False, byID=False, returnData=False, **kwargs):
        """Download the light curves for the retrieved GRBs.

        This function downloads the light curves of the GRBs found when
        this query was executed. Of course, this means you have to have
        executed the query first!

        The light curve data will be stored in the ``lightCurves``
        variable of this object, and optionally returned as well. The
        light curves files can also be downloaded directly from the
        website to disk.

        Parameters
        ----------

        subset : pandas.Series, optional
            A pandas series defining a subset of rows in the query
            results to download.

        byName : bool, optional
            Force the results to be indexed by the GRB name.

        byID : bool, optional
            Force the results to be indexed by targetID.
            Requires that column to have been retrieved by your query.

        returnData : bool, optional
            Whether the light curve data should be returned by this
            function, as well as saved in the "lightCurves"
            variable.

        **kwargs : dict
            Other arguments which are passed to
            ``ukssdc.data.GRB.getlightCurves()``.
            See its help for more information.

        """
        if not self.haveResults:
            raise RuntimeError("This query has not been executed, cannot download!")

        # Check if we are doing ID or name.
        data = {}
        (whichCol, whichArg) = self._handleArgs(byName=byName, byID=byID)

        # Now set up data, which will will pass as **data which will be received either as
        # sourceID = [...] or sourceName = [...]
        data[whichArg] = []

        # But before we can populate it, we may have to handle a subset:
        if subset is not None:
            if not isinstance(subset, pd.core.series.Series):
                raise ValueError("Subset parameter must be a pandas series")
            data[whichArg] = self._results.loc[subset][whichCol].tolist()
        else:
            data[whichArg] = self._results[whichCol].tolist()

        # if "returnData" not in kwargs:
        #     kwargs["returnData"] = True
        kwargs["returnData"] = True
        if "saveData" not in kwargs:
            kwargs["saveData"] = False

        tmp = dGRB.getLightCurves(
            silent=self.silent,
            verbose=self.verbose,
            **data,
            **kwargs,
        )

        if self._prodData["lightCurves"] is None:
            self._prodData["lightCurves"] = tmp
        else:
            # Can't just merge the dicts, annoyingly, because this is not a deep (recursive) merge, so will not
            # add extra bands. However, I don't need a full recurse, just at the source level, so it's not too hard.
            for i in tmp.keys():
                if i in self._prodData["lightCurves"]:
                    tmpDS = self._prodData["lightCurves"][i]["Datasets"]
                    tmpDS.extend(x for x in tmp[i]["Datasets"] if x not in tmpDS)
                    self._prodData["lightCurves"][i].update(tmp[i])
                    # self._prodData['lightCurves'][i] = {**self._prodData['lightCurves'][i], **tmp[i]}
                    self._prodData["lightCurves"][i]["Datasets"] = tmpDS
                else:
                    self._prodData["lightCurves"][i] = tmp[i]

        if returnData:
            return self._prodData["lightCurves"]

    def saveLightCurves(self, **kwargs):
        """Save the light curves to text files.

        This function is used to save the light curves that have been
        downloaded and are stored in the self.lightCurves
        variable of this GRBQuery object. These light curves
        can be saved as simple text file (with a user-settable
        column separator), or as files formatted for the qdp programme.

        All this function actually does it to call the
        saveLightCurves() in the related module:
        ``ukssdc.data.GRB``; further detail on the parameters that
        can be used are documented for that function.

        Parameters
        ----------

        **kwargs : dict
            A set of parameters to pass to the main save function in
            ``ukssdc.data.GRB``

        """
        # Check that we have downloaded some light curves!
        if self.lightCurves is None:
            raise RuntimeError("There are no light curves to save!")
        dGRB.saveLightCurves(self.lightCurves, silent=self.silent, verbose=self.verbose, **kwargs)

    def clearLightCurves(self):
        """Clear self.lightCurves"""
        self._prodData["lightCurves"] = None

    def getSpectra(self, subset=None, byName=False, byID=False, returnData=False, **kwargs):
        """Download the spectra for the retrieved GRBs.

        This function downloads the spectra of the GRBs found when this
        query was executed. Of course, this means you have to have
        executed the query first!

        The spectral data will be stored in the ``spectra``
        variable of this object, and optionally returned as well. The
        spectral files can also be downloaded directly from the
        website to disk.

        Parameters
        ----------

        subset : pandas.Series, optional
            A pandas series defining a subset of rows to download.

        byName : bool, optional
            Force the results to be indexed by IAUName.
            Requires that column to have been retrieved by your query.

        byID : bool, optional
            Force the results to be indexed by SXPS_ID.
            Requires that column to have been retrieved by your query.

        returnData : bool, optional
            Whether the light curve data should be returned by this
            function, as well as saved in the "lightCurves"
            variable.

        **kwargs : dict, optional
            Parameters to pass to ukssdc.data.GRB.getSpectra()

        """
        if not self.haveResults:
            raise RuntimeError("This query has not been executed, cannot download!")

        # Check if we are doing ID or name.
        data = {}
        (whichCol, whichArg) = self._handleArgs(byName=byName, byID=byID)

        # Now set up data, which will will pass as **data which will be received either as
        # sourceID = [...] or sourceName = [...]
        data[whichArg] = []

        # But before we can populate it, we may have to handle a subset:
        if subset is not None:
            if not isinstance(subset, pd.core.series.Series):
                raise ValueError("Subset parameter must be a pandas series")
            data[whichArg] = self._results.loc[subset][whichCol].tolist()
        else:
            data[whichArg] = self._results[whichCol].tolist()

        # if "returnData" not in kwargs:
        #     kwargs["returnData"] = True
        kwargs["returnData"] = True
        if "saveData" not in kwargs:
            kwargs["saveData"] = False
        if "saveImages" not in kwargs:
            kwargs["saveImages"] = False

        tmp = dGRB.getSpectra(
            silent=self.silent,
            verbose=self.verbose,
            **data,
            **kwargs,
        )

        if self._prodData["spectra"] is None:
            self._prodData["spectra"] = tmp
        else:
            self._prodData["spectra"].update(tmp)

        if returnData:
            return self._prodData["spectra"]

    def plotLightCurves(self, whichGRB, **kwargs):
        """Plot the light curves of a specific GRB.

        This is a wrapper to the ``ukssdc.plotLightCurve()`` function,
        giving you a plot of the light curve in question.

        You can only plot one GRB at a time (at present) and so you
        must provide the identifier (name or targetID) of a single
        downloaded GRB.

        Parameters
        ----------

        whichGRB : str
            The identifier of a GRB to download. Must be a key to the
            ``lightCurves`` dict.

        **kwargs : dict
            Arguments to pass to ``plotLightCurve()``

        """
        if self.lightCurves is None:
            raise RuntimeError("I don't have any light curves to plot!")

        if whichGRB not in self.lightCurves:
            raise RuntimeError(f"`{whichGRB}` is not a recognised, downloaded GRB.")

        if "silent" not in kwargs:
            kwargs["silent"] = self.silent
        if "verbose" not in kwargs:
            kwargs["verbose"] = self.verbose
        return lcPlot(self.lightCurves[whichGRB], **kwargs)

    def saveSpectra(self, **kwargs):
        """Save the spectra files to disk.

        This function is used to save the spectra that have been
        downloaded and are stored in the self.spectra
        variable of this GRBQuery object.

        All this function actually does it to call the
        saveSpectra() in the related module:
        ``ukssdc.data.GRB``; further detail on the parameters that
        can be used are documented for that function.

        Parameters
        ----------

        **kwargs : dict
            A set of parameters to pass to the main save function in
            ``ukssdc.data.GRB``

        """
        # Check that we have downloaded some light curves!
        if self.spectra is None:
            raise RuntimeError("There are no spectra to save!")

        dGRB.saveSpectra(self.spectra, silent=self.silent, verbose=self.verbose, **kwargs)

    def clearSpectra(self):
        """Clear self.spectra"""
        self._prodData["spectra"] = None

    def getBurstAnalyser(self, subset=None, byName=False, byID=False, returnData=False, **kwargs):
        """Download the burst analyser data for the retrieved GRBs.

        This function downloads the burst analyser data of the GRBs
        found when this query was executed. Of course, this means you
        have to have executed the query first!

        The burst analyser data will be stored in the ``burstAnalyser``
        variable of this object, and optionally returned as well. The
        data files can also be downloaded directly from the website to
        disk.

        Parameters
        ----------

        subset : pandas.Series, optional
            A pandas series defining a subset of rows to download.

        byName : bool, optional
            Force the results to be indexed by IAUName.
            Requires that column to have been retrieved by your query.

        byID : bool, optional
            Force the results to be indexed by SXPS_ID.
            Requires that column to have been retrieved by your query.

        returnData : bool, optional
            Whether the burst analyser data should be returned by this
            function, as well as saved in the "burstAnalyser"
            variable.

        **kwargs : dict, optional
            Parameters to pass to ukssdc.data.GRB.getBurstAnalyser()

        """
        if not self.haveResults:
            raise RuntimeError("This query has not been executed, cannot download!")

        # Check if we are doing ID or name.
        data = {}
        (whichCol, whichArg) = self._handleArgs(byName=byName, byID=byID)

        # Now set up data, which will will pass as **data which will be received either as
        # sourceID = [...] or sourceName = [...]
        data[whichArg] = []

        # But before we can populate it, we may have to handle a subset:
        if subset is not None:
            if not isinstance(subset, pd.core.series.Series):
                raise ValueError("Subset parameter must be a pandas series")
            data[whichArg] = self._results.loc[subset][whichCol].tolist()
        else:
            data[whichArg] = self._results[whichCol].tolist()

        # if "returnData" not in kwargs:
        #     kwargs["returnData"] = True
        kwargs["returnData"] = True
        if "saveData" not in kwargs:
            kwargs["saveData"] = False
        if "downloadTar" not in kwargs:
            kwargs["downloadTar"] = False

        tmp = dGRB.getBurstAnalyser(
            silent=self.silent,
            verbose=self.verbose,
            **data,
            **kwargs,
        )

        if self._prodData["burstAnalyser"] is None:
            self._prodData["burstAnalyser"] = tmp
        else:
            self._prodData["burstAnalyser"].update(tmp)

        if returnData:
            return self._prodData["burstAnalyser"]

    def saveBurstAnalyser(self, **kwargs):
        """Save the burst analyser data to disk.

        This function is used to save the burst analyser data that have
        been downloaded and are stored in the self.burstAnalyser
        variable of this GRBQuery object.

        All this function actually does it to call the
        saveBurstAnalyser() in the related module:
        ``ukssdc.data.GRB``; further detail on the parameters
        that can be used are documented for that function.

        Parameters
        ----------

        **kwargs : dict
            A set of parameters to pass to the ``saveBurstAnalyser()``
            function in ``ukssdc.data.GRB``.

        """
        # Check that we have downloaded some light curves!
        if self.burstAnalyser is None:
            raise RuntimeError("There are no burst analyser data to save!")

        dGRB.saveBurstAnalyser(self.burstAnalyser, silent=self.silent, verbose=self.verbose, **kwargs)

    def clearBurstAnalysers(self):
        """Clear self.burstAnalyser"""
        self._prodData["burstAnalyser"] = None

    def getObsData(self, subset=None, **kwargs):
        """Download the obs data for the retrieved GRBs.

        This function downloads the obs data of the GRBs found when
        this query was executed. Of course, this means you have to have
        executed the query first!

        It basically wraps around ``ukssdc.data.downloadObsData()`` and
        most of the arguments you supply are just passed to that.

        Parameters
        ----------

        subset : pandas.Series, optional
            A pandas series defining a subset of rows in the query
            results to download.

        **kwargs : dict
            Other arguments which are passed to
            ``ukssdc.data.downloadObsData()``.

        """
        if not self.haveResults:
            raise RuntimeError("This query has not been executed, cannot download!")

        whichCol = None
        whichArg = None

        if self._targetCol in self.results.columns:
            whichCol = self._targetCol
            whichArg = "targetID"
        elif self._nameCol in self.results.columns:
            whichCol = self._nameCol
            whichArg = "GRBName"
        else:
            raise RuntimeError("Cannot get data as neither a name or targetID have been retrieved")

        data = {}
        # But before we can populate it, we may have to handle a subset:
        if subset is not None:
            if not isinstance(subset, pd.core.series.Series):
                raise ValueError("Subset parameter must be a pandas series")
            data[whichArg] = self._results.loc[subset][whichCol].tolist()
        else:
            data[whichArg] = self._results[whichCol].tolist()

        dGRB.getObsData(silent=self.silent, verbose=self.verbose, **kwargs, **data)

    def getPositions(self, subset=None, byName=False, byID=False, returnData=False, **kwargs):
        """Download the positions for the retrieved GRBs.

        This function downloads the positions of the GRBs found when
        this query was executed. Of course, this means you have to have
        executed the query first!

        The position data will be stored in the ``positions``
        variable of this object, and optionally returned as well.

        Parameters
        ----------

        subset : pandas.Series, optional
            A pandas series defining a subset of rows in the query
            results to download.

        byName : bool, optional
            Force the results to be indexed by the GRB name.

        byID : bool, optional
            Force the results to be indexed by targetID.
            Requires that column to have been retrieved by your query.

        returnData : bool, optional
            Whether the position data should be returned by this
            function, as well as saved in the "position"
            variable.

        **kwargs : dict
            Other arguments which are passed to
            ``ukssdc.data.GRB.getPositions()``.
            See its help for more information.

        """
        if not self.haveResults:
            raise RuntimeError("This query has not been executed, cannot download!")

        # Check if we are doing ID or name.
        data = {}
        (whichCol, whichArg) = self._handleArgs(byName=byName, byID=byID)

        # Now set up data, which will will pass as **data which will be received either as
        # sourceID = [...] or sourceName = [...]
        data[whichArg] = []

        # But before we can populate it, we may have to handle a subset:
        if subset is not None:
            if not isinstance(subset, pd.core.series.Series):
                raise ValueError("Subset parameter must be a pandas series")
            data[whichArg] = self._results.loc[subset][whichCol].tolist()
        else:
            data[whichArg] = self._results[whichCol].tolist()

        tmp = dGRB.getPositions(
            silent=self.silent,
            verbose=self.verbose,
            **data,
            **kwargs,
        )

        if self._prodData["positions"] is None:
            self._prodData["positions"] = tmp
        else:
            self._prodData["positions"] = {**self._prodData["positions"], **tmp}

        if returnData:
            return tmp

    def clearPositions(self):
        """Clear self.positions"""
        self._prodData["positions"] = None



from .api_common import TOOAPI_Baseclass
from .api_skycoord import TOOAPI_SkyCoord
from .api_status import Swift_TOO_Status


class Swift_Resolve(TOOAPI_Baseclass, TOOAPI_SkyCoord):
    """Swift_Resolve class

    Performs name resolution using Simbad, TNS or MARS. Simply give the name of
    the source, and it will return `ra` and `dec` in decimal degrees, or a
    astropy SkyCoord (`skycoord`).

    Attributes
    ----------
    name : str
        name of the object to have coordinates resolved.
    username: str
        Swift TOO API username (default 'anonymous')
    shared_secret: str
        TOO API shared secret (default 'anonymous')
    ra : float
        Right Ascension of resolved target in J2000 (decimal degrees)
    dec : float
        Declination of resolved target in J2000 (decimal degrees)
    resolver : str
        Name of name resolving service used
    skycoord : SkyCoord
        SkyCoord version of RA/Dec if astropy is installed
    status : str
        status of API request
    """

    # API input and return values definition
    _parameters = ["username", "name"]
    _attributes = ["ra", "dec", "resolver", "status"]
    # Other API classes that may be used by this class
    _subclasses = [Swift_TOO_Status]
    # Local parameters
    _local = ["shared_secret"]
    # API name
    api_name = "Swift_Resolve"

    def __init__(self, *args, **kwargs):
        """
        Parameters
        ----------
        name : str
            name of the object to have coordinates resolved.
        username: str
            Swift TOO API username (default 'anonymous')
        shared_secret: str
            TOO API shared secret (default 'anonymous')
        """
        # Input parameters
        self.name = None
        self.username = "anonymous"
        # Returned parameters
        self.ra = None
        self.dec = None
        self.resolver = None
        # Initiate status class
        self.status = Swift_TOO_Status()
        # Parse argument keywords
        self._parseargs(*args, **kwargs)

        # See if we pass validation from the constructor, but don't record
        # errors if we don't
        if self.validate():
            self.submit()
        else:
            self.status.clear()

    def validate(self):
        """Validate API submission before submit

        Returns
        -------
        bool
            Was validation successful?
        """
        if self.name is not None and self.shared_secret is not None:
            return True
        else:
            return False

    @property
    def _table(self):
        """Displays values in class as a table"""
        if self.ra is not None:
            header = ["Name", "RA (J2000)", "Dec (J2000)", "Resolver"]
            table = [[self.name, f"{self.ra:.5f}", f"{self.dec:.5f}", self.resolver]]
            return header, table
        else:
            return [], []


class TOOAPI_AutoResolve:
    """Mixin to automatically any given `name` into RA/Dec coordinates using
    `Swift_Resolve`"""

    _name = None
    _source_name = None
    resolve = None

    @property
    def name(self):
        return self._name

    @property
    def source_name(self):
        return self._source_name

    def __name_setter(self, sourcename):
        """If you set a name, use `Swift_Resolve` to retrieve it's `ra` and `dec`."""
        if self._name != sourcename:
            self._name = sourcename
            self._source_name = sourcename
            self.resolve = Swift_Resolve(name=sourcename)
            if self.resolve.ra is not None:
                self.ra = self.resolve.ra
                self.dec = self.resolve.dec
            else:
                self.status.error("Could not resolve name.")

    @name.setter
    def name(self, sourcename):
        self.__name_setter(sourcename)

    @source_name.setter
    def source_name(self, sourcename):
        self.__name_setter(sourcename)


# Shorthand alias for class
Resolve = Swift_Resolve



from datetime import timedelta

from .api_common import TOOAPI_Baseclass
from .api_daterange import TOOAPI_Daterange
from .api_resolve import TOOAPI_AutoResolve
from .api_skycoord import TOOAPI_SkyCoord
from .api_status import TOOStatus
from .swift_clock import TOOAPI_ClockCorrect
from .swift_data import TOOAPI_DownloadData
from .swift_instruments import TOOAPI_Instruments
from .swift_obsid import TOOAPI_ObsID
from .swift_obsquery import Swift_Observation, Swift_Observations


class Swift_PPSTEntry(
    TOOAPI_Baseclass,
    TOOAPI_SkyCoord,
    TOOAPI_ObsID,
    TOOAPI_DownloadData,
    TOOAPI_Instruments,
    TOOAPI_ClockCorrect,
):
    """Class that defines an individual entry in the Swift Pre-Planned Science
    Timeline

    Attributes
    ----------
    begin : datetime
        begin time of observation
    end : datetime
        end time of observation
    targetid : int
        target ID  of the observation
    seg : int
        segment number of the observation
    xrt : str
        XRT mode of the observation
    uvot : str
        Hex string UVOT mode of the observation
    bat : int
        BAT mode of the observation
    exposure : timedelta
        exposure time of the observation
    ra : float
        Right Ascension of pointing in J2000 (decimal degrees)
    dec : float
        Declination of pointing in J2000 (decimal degrees)
    roll : float
        roll angle of the observation (decimal degrees)
    skycoord : SkyCoord
        SkyCoord version of RA/Dec if astropy is installed
    targname : str
        Target name of the primary target of the observation
    """

    # Core API defs
    _parameters = [
        "begin",
        "end",
        "targname",
        "ra",
        "dec",
        "roll",
        "targetid",
        "seg",
        "xrt",
        "uvot",
        "bat",
        "fom",
    ]
    _varnames = {
        "begin": "Begin Time",
        "end": "End Time",
        "targname": "Target Name",
        "ra": "RA(J2000)",
        "dec": "Dec(J200)",
        "roll": "Roll (deg)",
        "targetid": "Target ID",
        "seg": "Segment",
        "xrt": "XRT Mode",
        "uvot": "UVOT Mode",
        "bat": "BAT Mode",
        "fom": "Figure of Merit",
        "obsnum": "Observation Number",
        "exposure": "Exposure (s)",
        "slewtime": "Slewtime (s)",
    }
    _attributes = []
    _subclasses = [TOOStatus]
    # API name
    api_name = "Swift_PPST_Entry"

    def __init__(self):
        # Parameters
        self.begin = None
        self.end = None
        self.ra = None
        self.dec = None
        self.roll = None
        self.targname = None
        self.targetid = None
        self.seg = None
        self.slewtime = timedelta(0)  # Slewtime isn't reported in plans

        # Swift_PPST returns a bunch of stuff we don't care about, so just take the things we do
        self.ignorekeys = True

    @property
    def exposure(self):
        return self.end - self.begin

    @property
    def _table(self):
        _parameters = ["begin", "end", "targname", "obsnum", "exposure"]
        header = [self._header_title(row) for row in _parameters]
        return header, [[self.begin, self.end, self.targname, self.obsnum, self.exposure.seconds]]


class Swift_PPST(
    TOOAPI_Baseclass,
    TOOAPI_Daterange,
    TOOAPI_SkyCoord,
    TOOAPI_ObsID,
    TOOAPI_DownloadData,
    TOOAPI_AutoResolve,
    TOOAPI_ClockCorrect,
):
    """Class to fetch Swift Pre-Planned Science Timeline (PPST) for given
    constraints. Essentially this will return what Swift was planned to observe
    and when, for given constraints. Constraints can be for give coordinate
    (SkyCoord or J2000 RA/Dec) and radius (in degrees), a given date range, or a
    given target ID (targetid) or Observation ID (obsnum).

    Attributes
    ----------
    begin : datetime
        begin time of visibility window
    end : datetime
        end time of visibility window
    length : timedelta
        length of visibility window
    ra : float
        Right Ascension of target in J2000 (decimal degrees)
    dec : float
        Declination of target in J2000 (decimal degrees)
    skycoord : SkyCoord
        SkyCoord version of RA/Dec if astropy is installed
    username: str
        Swift TOO API username (default 'anonymous')
    shared_secret: str
        TOO API shared secret (default 'anonymous')
    entries : list
        List of observations (`Swift_AFSTEntry`)
    status : TOOStatus
        Status of API request
    ppstmax: datetime
        When is the PPST valid up to
    """

    # Core API definitions
    _parameters = [
        "username",
        "begin",
        "end",
        "ra",
        "dec",
        "radius",
        "targetid",
        "obsnum",
    ]
    _attributes = ["status", "ppstmax", "entries"]
    _local = ["obsid", "name", "skycoord", "length", "target_id", "shared_secret"]
    _subclasses = [Swift_PPSTEntry, TOOStatus]
    api_name = "Swift_PPST"

    def __init__(self, *args, **kwargs):
        """
        Parameters
        ----------
        begin : datetime
            begin time of window
        end : datetime
            end time of window
        ra : float
            Right Ascension of target in J2000 (decimal degrees)
        dec : float
            Declination of target in J2000 (decimal degrees)
        radius : float
            radius in degrees to search around (default 0.197)
        targetid : int
            target ID of target
        obsid : int / str
            Observation ID of target, either in spacecraft (int) or SDC (str)
            formats
        skycoord : SkyCoord
            SkyCoord version of RA/Dec if astropy is installed
        username : str
            username for TOO API (default 'anonymous')
        length : timedelta
            length of window
        shared_secret : str
            shared secret for TOO API (default 'anonymous')
        """
        # Coordinate search
        self.ra = None
        self.dec = None
        self.radius = 11.8 / 60  # Default 11.8 arcmin - XRT FOV
        # begin and end boundaries
        self.begin = None
        self.end = None
        self.length = None
        # Search on targetid/obsnum
        self.targetid = None
        self.obsnum = None
        # Login
        self.username = "anonymous"
        # PPST entries go here
        self.entries = list()
        # Status of request
        self.status = TOOStatus()
        # Latest PPST
        self.ppstmax = None
        # Observations
        self._observations = Swift_Observations()

        # Parse argument keywords
        self._parseargs(*args, **kwargs)

        # See if we pass validation from the constructor, but don't record
        # errors if we don't
        if self.validate():
            self.submit()
        else:
            self.status.clear()

    @property
    def _table(self):
        if len(self.entries) > 0:
            header = self.entries[0]._table[0]
        else:
            header = []
        return header, [ppt._table[1][0] for ppt in self]

    @property
    def observations(self):
        if len(self.entries) > 0 and len(self._observations.keys()) == 0:
            for q in self.entries:
                self._observations[q.obsnum] = Swift_Observation()
            _ = [self._observations[q.obsnum].append(q) for q in self.entries]
        return self._observations

    def __getitem__(self, index):
        return self.entries[index]

    def __len__(self):
        return len(self.entries)

    def validate(self):
        """Validate API submission before submit

        Returns
        -------
        bool
            Was validation successful?
        """
        # How many search keys? Require at least one
        keys = self.api_data.keys()

        # We need at least one of these keys to be submitted
        req_keys = ["begin", "ra", "dec", "targetid", "obsnum"]

        # Check how many of them are in the request
        total_keys = 0
        for key in keys:
            if key in req_keys:
                if self.api_data[key]:
                    total_keys += 1

        # We need at least one key to be set
        if total_keys == 0:
            self.status.error("Please supply search parameters to narrow search.")
            return False

        # Check if ra or dec are in keys, we have both.
        if "ra" in keys or "dec" in keys:
            if not ("ra" in keys and "dec" in keys):
                self.status.error("Must supply both RA and Dec.")
                return False

        return True


# Class aliases for better PEP8 compliant and future compat
Swift_PlanQuery = Swift_PPST
PlanQuery = Swift_PPST
PPST = Swift_PPST
PPSTEntry = Swift_PPSTEntry
Swift_PPST_Entry = Swift_PPSTEntry



from datetime import datetime

from .api_common import TOOAPI_Baseclass, convert_to_dt, swiftdatetime
from .api_status import TOOStatus


class Swift_Clock(TOOAPI_Baseclass):
    """Class to obtain clock corrections, MET values and corrected UTC times for
    Swift. Typical use of the class is to pass a MET value, Swift Time
    (essentially MET in datetime format), or a UTC datetime. The API returns
    back, all three values, corrected for Swift's clock drift and leap seconds.

    Attributes
    ----------
    met : int
        Mission Elapsed Time (seconds). Amount of seconds since 2001-01-01
        00:00:00 UT. Note that MET is measured using Swift's internal clock, and
        are not corrected for clock drift.
    swifttime : datetime
        Spacecraft time converted to datetime units. Note that dates are note
        corrected for leap seconds or clock drift.
    utctime : datetime
        Universal Time value, corrected for Swift's clock drift and leap
        seconds.
    utcf : float
        Univeral Time Correction Factor. The number of seconds to apply to
        correct MET / Spacecraft time into UTC.
    status : TOOStatus
        Status of API request
    username : str (default 'anonymous')
        TOO API username.
    """

    # API details
    api_name = "Swift_Clock"
    # Arguments
    _parameters = ["username", "met", "swifttime", "utctime"]
    # Returned values
    _attributes = ["status", "entries"]
    # Returned classes
    _subclasses = [TOOStatus, swiftdatetime]
    # Local parameters
    _local = ["shared_secret"]

    def __init__(self, *args, **kwargs):
        """
        Parameters
        ----------
        met : int
            Mission Elapsed Time (seconds). Amount of seconds since 2001-01-01
            00:00:00 UT. Note that MET is measured using Swift's internal clock, and
            are not corrected for clock drift.
        swifttime : datetime
            Spacecraft time converted to datetime units. Note that dates are note
            corrected for leap seconds or clock drift.
        utctime : datetime
            Universal Time value, corrected for Swift's clock drift and leap
            seconds.
        username : str (default 'anonymous')
            TOO API username.
        shared_secret : str (default 'anonymous')
            TOO API shared secret
        """
        # attributes
        self._met = None
        self._swifttime = None
        self._utctime = None
        self._utcf = None
        # parse arguments
        self._parseargs(*args, **kwargs)
        self.status = TOOStatus()
        self.entries = None

        # Submit if enough parameters are passed to the constructor
        if self.validate():
            self.submit()
        else:
            self.status.clear()

    @property
    def met(self):
        if self._met is None and self.entries is not None:
            self._met = [e.met for e in self]
            if len(self._met) == 1:
                self._met = self._met[0]
        return self._met

    @met.setter
    def met(self, mettime):
        self._met = mettime

    @property
    def utcf(self):
        if self._utcf is None and self.entries is not None:
            self._utcf = [e.utcf for e in self]
            if len(self._utcf) == 1:
                self._utcf = self._utcf[0]
        return self._utcf

    @utcf.setter
    def utcf(self, utcftime):
        self._utcf = utcftime

    @property
    def utctime(self):
        if self._utctime is None and self.entries is not None:
            self._utctime = [entry.utctime for entry in self.entries]
            if len(self._utctime) == 1:
                self._utctime = self._utctime[0]
        return self._utctime

    @utctime.setter
    def utctime(self, utct):
        if type(utct) == swiftdatetime:
            if utct.utctime is None:
                raise TypeError("swiftdatetime does not have utctime set")
            else:
                self._utctime = utct.utctime
        else:
            self._utctime = self.__todt(utct)

    @property
    def swifttime(self):
        if self._swifttime is None and self.entries is not None:
            self._swifttime = [e.swifttime for e in self]
            if len(self._swifttime) == 1:
                self._swifttime = self._swifttime[0]
        return self._swifttime

    @swifttime.setter
    def swifttime(self, swiftt):
        if type(swiftt) == swiftdatetime:
            if swiftt.swifttime is None:
                raise TypeError("swiftdatetime does not have swifttime set")
            else:
                self._swifttime = swiftt.swifttime
        else:
            self._swifttime = self.__todt(swiftt)
        self._swifttime = self.__todt(swiftt)

    def __todt(self, dt):
        """Simple method to revert a swiftdatetime back to a datetime"""
        if isinstance(dt, (list, tuple)):
            return [convert_to_dt(d) for d in dt]
        else:
            return convert_to_dt(dt)

    def __getitem__(self, index):
        return self.entries[index]

    def __len__(self):
        if self.entries is not None:
            return len(self.entries)
        else:
            return 0

    @property
    def _table(self):
        if len(self) > 0:
            values = [self[i]._table[1][0] for i in range(len(self))]
            header = self[0]._table[0]
        else:
            header, values = [], []
        return header, values

    def validate(self):
        """Validate API submission before submit

        Returns
        -------
        bool
            Was validation successful?
        """
        # Check what the source is, and set _isutc appropriately
        if self.swifttime is not None or self.met is not None:
            self._isutc = False
        else:
            self._isutc = True
        # Basic check that anything is set
        if self.met is not None or self.swifttime is not None or self.utctime is not None:
            return True

    def to_utctime(self):
        """Convert all entries to a UTC time base"""
        mets = [entry.met for entry in self.entries]
        utcfs = [entry.utcf for entry in self.entries]
        self.entries = [swiftdatetime.frommet(mets[i], utcf=utcfs[i], isutc=True) for i in range(len(mets))]

    def to_swifttime(self):
        """Convert all entries to a Swift Time base"""
        mets = [entry.met for entry in self.entries]
        utcfs = [entry.utcf for entry in self.entries]
        self.entries = [swiftdatetime.frommet(mets[i], utcf=utcfs[i], isutc=False) for i in range(len(mets))]

    # Aliases
    mettime = met
    swift = swifttime
    utc = utctime


Clock = Swift_Clock


def index_datetimes(dictionary, i=0, values=[], setvals=None):
    """Recursively spider a dictionary looking for datetimes and updating them
    if necessary"""
    # Don't spider internal variables
    keys = [key for key in list(dictionary.keys())]
    # Go through all keys
    for key in keys:
        value = dictionary[key]
        # Don't index any `Swift_Clock`s
        if type(value) is Swift_Clock:
            continue
        # If value is another dict, recurse
        if isinstance(value, dict):
            i, values = index_datetimes(value, i, values, setvals=setvals)

        # If value is a list, recurse one by one
        elif isinstance(value, (list, tuple)):
            for j in range(len(value)):
                i, values = index_datetimes({f"value{j}": value[j]}, i, values, setvals=setvals)

        # If value is a datetime, record and/or update to the result from
        # Swift_Clock, increment the counter
        elif isinstance(value, datetime):
            if setvals is not None:
                dictionary[key] = setvals[i]
            values.append(value)
            i += 1

        # If value is a class object other than a datetime, convert to dict and recurse
        elif hasattr(value, "__dict__"):
            i, values = index_datetimes(value.__dict__, i, values, setvals=setvals)

    # Return counter and values
    return i, values


class TOOAPI_ClockCorrect:
    """Mixin for clock correction. Provides the  `clock_correct` method, which
    spiders through a class looking for datetimes, submits them to Swift_Clock,
    and then replaces them all with the results of Swift_Clock."""

    def clock_correct(self):
        """Spider through the class dictionary recording datetimes, and then
        updating them using Swift_Clock"""
        if not hasattr(self, "_clock"):
            # Read in all datetime values into an array
            _, datevalues = index_datetimes(self.__dict__, 0, [])

            # What is the base time format for this class? Send that to Clock for
            # clock correction
            dts = [dt for dt in datevalues]
            if len(dts) > 0:
                if self._isutc:
                    self._clock = Swift_Clock(utctime=dts)
                else:
                    self._clock = Swift_Clock(swifttime=dts)

                # Replace existing datetime values with clock corrected swiftdatetimes
                _, _ = index_datetimes(self.__dict__, 0, [], setvals=self._clock)

        # After clock correction, make UTC the default time system
        self.to_utctime()

    def to_utctime(self):
        """Convert times to a UTC base"""
        self._clock.to_utctime()
        _, _ = index_datetimes(self.__dict__, 0, [], setvals=self._clock)

    def to_swifttime(self):
        """Convert times to a Swift time base"""
        self._clock.to_swifttime()
        _, _ = index_datetimes(self.__dict__, 0, [], setvals=self._clock)

    def _header_title(self, parameter):
        """Add UTC or Swift to headers in table depending on the default"""
        title = self._varnames[parameter]
        value = getattr(self, parameter)
        if type(value) == swiftdatetime:
            if value.isutc:
                title += " (UTC)"
            else:
                title += " (Swift)"
        return title



"""Data download functions.

This module provides functions for downloading data or products.

"""

__docformat__ = "restructedtext en"


import os
import os.path
import io
import shutil
import requests
import re
import fnmatch
import pandas as pd
import subprocess
from .. import main as base


_allowedInstruments = ["bat", "xrt", "uvot"]
_allowedSources = ["uk", "uk_reproc", "us", "italy"]

try:
    from tqdm.auto import tqdm
except ImportError:

    def tqdm(*args, **kwargs):
        """A real simple replacement for tqdm if it's not locally installed"""
        print(f"Downloading {len(args[0])} files...")
        return args[0]


def downloadObsData(
    obsid,
    source="uk_reproc",
    instruments="all",
    destDir="data",
    getAuxil=True,
    getTDRSS=False,
    getLog=False,
    match=None,
    reMatch=None,
    clobber=False,
    skipErrors=False,
    silent=True,
    verbose=False,
    **kwargs,
):
    """Download observation data files.

    This function receives an obsid, or a list thereof, and
    downloads the Swift obsdata for this. It will create one directory
    per obsID, and download files into that. If the directory exists,
    an error will be thrown unless ``clobber=True`` was set, in which
    case the directory will be removed and recreated.

    Parameters
    ----------

    obsid : str or int or list or tuple
        The obsid to download.

    source : str
        Which datacentre to downloaded the data from.

    instruments : str or list or tuple.
        Which instruments to download data for. A list or 'all'.

    destDir : str, optional
        The top-level directory to save files into (default: 'data').

    getAuxil : bool, optional
        Whether to include the 'auxil' directory in the download
        (default: ``True``).

    getTDRSS : bool, optional
        Whether to include the 'tdrss' directory in the download
        (default: ``True``).

    getLog : bool, optional
        Whether to include the 'log' directory in the download
        (default: ``True``).

    match : str or list, optional
        A "glob"-style wildcard string to match against filenames; only
        files matching this will be downloaded. If a list is supplied
        then files which match any of the globs in the list will be
        downloaded. Note, only one of 'match' and 'reMatch' can be set.

    reMatch : str or list, optional
        A regular expression string to match against filenames; only
        files matching this will be downloaded. If a list is supplied
        then files which match any of the expressions in the list will
        be downloaded. Note, only one of 'match' and 'reMatch' can be
        set.

    clobber : bool, optional
        Whether to overwrite files if they exist (default: ``False``).

    skipErrors : bool, optional
        Whether to continue to the next file/observation if an error is
        encountered (default: ``False``).

    silent : bool, optional
        Whether to suppress all output (default: ``True``).

    verbose : bool, optional
        Whether to write verbose output (default: ``False``).

    """
    if verbose:
        silent = False

    # Check validity
    if isinstance(obsid, (str, int)):
        obsid = [obsid]

    if not isinstance(obsid, (list, tuple)):
        raise ValueError("`obsid` must be a list, tuple, int or string.")

    # -------------------------------------
    # Check instruments and handle which we want to get.
    dirs = None

    # Check that 'instruments' and 'source' are valid, and destDir exists
    if isinstance(instruments, str):
        if instruments.lower() == "all":
            dirs = _allowedInstruments
        else:
            raise ValueError("instruments parameter must be 'all' or a list")
    elif isinstance(instruments, (list, tuple)):
        dirs = []
        for i in instruments:
            if i.lower() not in _allowedInstruments:
                raise ValueError(f"`{i}` is not a recognised instrument")
            else:
                dirs.append(i.lower())
    elif instruments is None:
        dirs = []
    else:
        raise ValueError("instruments parameter must be 'all', None or a list")

    if getAuxil:
        dirs.append("auxil")
    if getTDRSS:
        dirs.append("tdrss")
    if getLog:
        dirs.append("log")

    # ---------------------------------------
    # Check match and rematch
    if (match is not None) and (reMatch is not None):
        raise ValueError("Cannot have both match and reMatch set")

    if match is not None:
        if isinstance(match, str):
            match = [match]
        if not isinstance(match, (list, tuple)):
            raise ValueError("`match` must be a list, tuple or string")
    if reMatch is not None:
        if isinstance(reMatch, str):
            reMatch = [reMatch]
        if not isinstance(reMatch, (list, tuple)):
            raise ValueError("`reMatch` must be a list, tuple or string")
        tmp = []
        for r in reMatch:
            tmp.append(re.compile(r))
        reMatch = tmp

    # ---------------------------------------
    # Is the source valid
    if source.lower() not in _allowedSources:
        raise ValueError(f"`{source}` is not a valid data source.")
    source = source.lower()

    # clobber = False
    # if "clobber" in kwargs.keys():
    #     clobber = bool(kwargs["clobber"])

    # skipErrors = False
    # if "skipErrors" in kwargs.keys():
    #     skipErrors = bool(kwargs["skipErrors"])

    # Do we need to make the dest dir?
    base._createDir(destDir, silent=silent, verbose=verbose)

    if not silent:
        print(f"Downloading {len(obsid)} datasets")

    for obs in obsid:
        if isinstance(obs, str):
            obs = int(obs)
        if isinstance(obs, int):
            obs = f"{obs:011d}"
        else:
            raise ValueError(f"Obsid `{obs}` is not valid.")
        if verbose:
            print(f"Getting obsid {obs}")

        # Path already exists?
        obsPath = f"{destDir}/{obs}"
        if os.path.exists(obsPath):
            if clobber:
                if not silent:
                    print(f"WARNING: Deleting / overwriting directory {obsPath}.")
                shutil.rmtree(obsPath)
            else:
                raise RuntimeError(f"Directory {obsPath} exists and clobber is False.")

        # Make the path
        base._createDir(obsPath, silent=silent, verbose=verbose)

        # Get the file list for this obs.
        fileData = _getFileList(obs, dirs, source, silent, verbose)
        fileTree = fileData["fileList"]

        # Does it exist?
        if ("dirs" not in fileTree.keys()) or ("files" not in fileTree.keys()):
            if skipErrors:
                if not silent:
                    print(f"Cannot find obs {obs}, skipping")
            else:
                raise RuntimeError(f"Cannot find obs {obs}.")

        if len(fileTree["files"]) == 0:
            if not silent:
                print(f"Obs `{obs}` has no files to download, skipping")
            continue

        # Make all directories
        for dir in fileTree["dirs"]:
            # Skip the top level dir, already made
            if len(dir) == 0:
                continue
            outDirName = f"{obsPath}/{dir}"
            base._createDir(outDirName, silent=silent, verbose=verbose)

        # Now get the files, use tqdm to plot progress:
        # display = not silent
        urlBase = fileData["url"]

        if silent:
            myList = fileTree["files"]
        else:
            myList = tqdm(fileTree["files"], desc=f"Downloading {obs}", unit="files")

        # for f in tqdm(fileTree["files"], desc=f"Downloading {obs}", unit="files", display=display):
        for f in myList:
            url = f"{urlBase}/{f}"
            outPath = f"{obsPath}/{f}"

            if match is not None:
                isOK = False
                for m in match:
                    if fnmatch.fnmatch(os.path.basename(f), m):
                        isOK = True
                        break
                if not isOK:
                    if verbose:
                        print(f"Skipping file {f}")
                    continue
            elif reMatch is not None:
                isOK = False
                for m in reMatch:
                    if m.search(os.path.basename(f)) is not None:
                        isOK = True
                        break
                if not isOK:
                    if verbose:
                        print(f"Skipping file {f}")
                    continue

            r = requests.get(url, stream=True, allow_redirects=True)
            if r.ok:
                filedata = r.raw.read()
                with open(outPath, "wb") as outfile:
                    outfile.write(filedata)
            else:
                if not skipErrors:
                    raise RuntimeError(f"Failed to download {url}")
                if not silent:
                    print(f"Failed to download {url}")


def downloadObsDataByTarget(targetID, silent=True, verbose=False, **kwargs):
    """Download observation data files by target.

    This function receives a targetID, or a list of targetIDs, and then
    calls ``downloadObsData()`` for each obsID within each target.

    Parameters
    ----------

    targetID : str or int or list or tuple
            The targetID/s to download.

    silent : bool, optional
        Whether to suppress all output (default: ``True``).

    verbose : bool, optional
        Whether to write verbose output (default: ``False``).

    **kwargs : dict, optional
        Arguments to pass to ``downloadObsData()``.

    """
    if verbose:
        silent = False

    # Check validity
    if isinstance(targetID, (str, int)):
        targetID = [targetID]

    if not isinstance(targetID, (list, tuple)):
        raise ValueError("`targetID` must be a list, tuple, int or string.")

    for targ in targetID:
        o = _getObsList(targ, silent=silent, verbose=verbose)
        if verbose:
            print(f"For {targ}, have to get obsIDs: {o}")
        downloadObsData(o, silent=silent, verbose=verbose, **kwargs)


def _getObsList(targetID, silent=True, verbose=False):
    """Internal function get the the obsID list for a targetID.

    Parameters
    ----------

    targetID : int or str
        The targetID.

    silent : bool, optional
        Whether to suppress all output (default: ``True``).

    verbose : bool, optional
        Whether to write verbose output (default: ``False``).

    Returns
    -------

    list
        The list of obsIDs.

    """
    if verbose:
        silent = False  # noqa

    sendData = {"targetID": targetID}
    ret = base.submitAPICall("getObsByTarg", sendData, minKeys=["obsList"], verbose=verbose)
    return ret["obsList"]


# --------------------------------------------------------------------
# Functions not explicitly exported to the public, although usable
# internally.


def _saveURLToFile(url, path, prefix=None, name=None, clobber=False, silent=True, verbose=False):
    """Internal function to download and save a file.

    Parameters
    ----------

    url : str
        The URL of the file to download.

    path : str
        The path to the directory into which the file should be saved.

    prefix : str, optional
        A prefix to prepend to file names

    name : str,optional
        The filename to save. If None uses the basename of the URL
    clobber: bool, optional
        Overwrite existing files? (default ``False``)

    silent : bool, optional
        Whether to suppress all output (default: ``True``).

    verbose : bool, optional
        Whether to write verbose output (default: ``False``).

    """
    if verbose:
        silent = False

    if prefix is None:
        prefix = ""
    if name is None:
        fname = os.path.basename(url)
    else:
        fname = name
    fname = f"{path}/{prefix}{fname}"
    if (os.path.exists(fname)) and (not clobber):
        if not silent:
            print(f"`{fname}` exists and clobber=False, SKIPPING")
        return False

    if verbose:
        print(f"Downloading file `{fname}`")

    # # TEMP LINES
    # from requests.auth import HTTPBasicAuth

    d = requests.get(
        url,
        # auth=HTTPBasicAuth("LSXPS", "CatPreview"),
    )
    if verbose:
        print(f"Saving file `{fname}`")

    with open(fname, "wb") as f:
        f.write(d.content)
        return True


def _getFileList(obs, dirs, source, silent=True, verbose=False):
    """Run an API call to get the file list.

    Parameters
    ----------

    obs : str
        The obsid.

    dirs : list
        The directories to list.

    source : str
        Where to get the data from.

    silent : bool, optional
        Whether to suppress all output (default: ``True``).

    verbose : bool, optional
        Whether to write lots of output (default: ``False``).

    Returns
    -------

    dict
        Containing the URL, directory and file list.

    """
    if verbose:
        silent = False  # noqa

    #
    sendData = {"obsid": obs, "source": source, "dirs": dirs}
    if verbose:
        print(f"Getting filelist for {obs} from {source} archive.")

    ret = base.submitAPICall("listObsFiles", sendData, minKeys=["url", "fileList"], verbose=verbose)
    return ret


def _handleLightCurve(data, oldCols=False, silent=True, verbose=False):
    """Convert light curves returned via API into pandas DataFrames.

    This is a generic function which receives light curves returned by
    the API, which have a fixed form, and converts them into pandas
    DataFrames for ease of use.

    Parameters
    ----------

    data : dict
        The data decoded from JSON.

    oldCols : bool, optional
        Whether to accept the old xrt_prods UL column
        (default: ``False``).

    silent : bool, optional
        Whether to suppress all output (default: ``True``).

    verbose : bool, optional
        Whether to write lots of output (default: ``False``).

    Returns
    -------
    pandas.DataFrame
        Light curve in a DataFrame

    """
    if verbose:
        silent = False  # noqa

    ret = {}
    url = {}
    skipKeys = []
    for key in data["Datasets"]:
        tmpKey = f"{key}Data"
        if tmpKey not in data:
            raise RuntimeError(f"Expected to find `{key}` data, but it is not present.")
        if "columns" not in data[tmpKey]:
            raise RuntimeError(f"`{key}` contains no column information.")
        cols = data[tmpKey]["columns"]
        skipKeys.append(tmpKey)

        if len(data[tmpKey]["data"]) > 0:
            if len(cols) != len(data[tmpKey]["data"][0]):
                # print(f"ARSE - {tmpKey}")
                # print(cols)
                # print(data[tmpKey]["data"])
                raise RuntimeError(f"Unable to handle the {tmpKey} light curve, corrupt data?")

        if "UL" in key and not oldCols:
            cols = ["UpperLimit" if x == "Rate" else x for x in cols]

        tmpDF = pd.DataFrame(data[tmpKey]["data"], columns=cols, dtype=str)

        types = {}
        for c in cols:
            if c in ("ObsID", "URL"):
                types[c] = str
            else:
                types[c] = float

        ret[key] = tmpDF.astype(types)
        # -- OPTION - could remove the RatePos and RateNeg columns if "UL" in tmpKey

        if "ObsID" in ret[key].columns:
            ret[key]["ObsID"] = ret[key]["ObsID"].apply(lambda x: f"{int(float(str(x).replace('::ObsID=', ''))):011d}")
        if "URL" in data[tmpKey]:
            url[key] = data[tmpKey]["URL"]

    for k, v in data.items():
        if k not in skipKeys:
            ret[k] = v

    if len(url) > 0:
        ret["URLs"] = url

    return ret


def _saveLightCurveFromURL(
    data, destDir="lc", prefix="", nosys="no", incbad="yes", clobber=False, skipErrors=False, silent=True, verbose=False
):
    """Save light curves to disk.

    This (internal) function receives a data structure, ideally produced
    by _handleLightCurve(), but at least containing the ``Datasets`` and
    ``URLs`` entries.

    It then downloads the relevant data, and saves them.

    Parameters
    ----------

    data : dict
        The light cure data

    destDir : str, optional
        The directory in which to save the light curves (default: "lc").

    prefix : str, optional
        A prefix to prepend to the file names before saving.

    nosys : str
        Whether to return the light curves from which the WT
        systematic error has been removed. Can be 'yes', 'no',
        'both' -- the latter returning datasets with and without the
        systematic error (default: 'no').

    incbad : str
        Whether to return the light curves in which the datapoints
        flagged as potentially unreliable to due centroiding issues
        have been included. Can be 'yes', 'no', 'both' -- the latter
        returning datasets with and without the unreliable
        datapoints (default: 'no').

    clobber : bool, optional
        Whether to overwrite files that already exist
        (default: ``False``).

    skipErrors : bool, optional
        Whether to continue if a problem occurs with one file
        (default: ``False``).

    silent : bool, optional
        Whether to suppress all output (default: ``True``).

    verbose : bool, optional
        Whether to write verbose output (default: ``False``).

    Returns
    -------

    int or None
        On success, the integer targetID; if the GRB is not found,
        ``None``

    """
    if verbose:
        silent = False

    base._createDir(destDir, silent=silent, verbose=verbose)

    if not (("URLs" in data) and ("Datasets" in data)):
        raise ValueError("data must contain keys 'URLs' and 'Datasets'.")

    # Now build the file list. This will be "fun", thanks to incbad, nosys, and the HRs sharing a file
    files = []
    for tmpkey in data["Datasets"]:
        if tmpkey not in data["URLs"]:
            raise RuntimeError(f"Cannot find `{tmpkey}` in the URLs. Is your swittools version up to date?")
        url = data["URLs"][tmpkey]
        # Skip if we have it:
        if url in files:
            continue

        # Do we want this file?
        if ("nosys" in tmpkey) and (nosys == "no"):
            continue
        if ("incbad" in tmpkey) and (incbad == "no"):
            continue
        if ("nosys" not in tmpkey) and (nosys == "yes"):
            continue
        if ("incbad" not in tmpkey) and (incbad == "yes"):
            continue
        files.append(url)

    if len(files) == 0:
        if not silent:
            print("No files selected to download")
        return

    # display = not silent
    if silent:
        myList = files
    else:
        myList = tqdm(files, desc="Downloading light curves", unit="files")

    #    for url in tqdm(files, desc="Downloading light curves", unit="files", display=display):
    for url in myList:
        file = os.path.basename(url)
        outPath = f"{destDir}/{prefix}{file}"
        if (os.path.exists(outPath)) and (not clobber):
            if not skipErrors:
                raise RuntimeError(f"File `{outPath}` already exists.")
            if not silent:
                print(f"File `{outPath}` already exists, SKIPPING")

        if verbose:
            print(f"Saving {outPath}")
        r = requests.get(url, stream=True, allow_redirects=True)
        if r.ok:
            filedata = r.raw.read()
            with open(outPath, "wb") as outfile:
                outfile.write(filedata)
        else:
            if not skipErrors:
                raise RuntimeError(f"Failed to download {url}")
            if not silent:
                print(f"Failed to download {url}")


def _getQDPHeader(data, curve, sep):
    """Get the qdp-style header lines for a given light curve.

    This is an internal function, not designed for user calls.

    This examines the DataFrame corresponding to a specific light
    curve, and works out what the READ commands for qdp should be.
    It then returns a string containing this line/s and also a
    qdp comment line giving the headers.

    Parameters
    ----------

    data : dict
        A light curve dict, created by this API.

    source : str
        The source being saved; must be an index in ``data``.

    curve : str
        The label defining the light curve; must be an index in
        ``data``.

    sep : str
        The column separator to use when writing the header.

    Returns
    -------

    str
        The header that needs printing before the data.

    """
    # I want this function to be fairly generic in theory, so that
    # we can support things with symmetric errors as well as asymmetric,
    # but there is no point wasting time working this out now, because
    # at present, the back end ONLY returns asymmetric time errors
    # and either asymmetric errors or no errors.
    # I don't see this changing in the foreseeable future, because
    # we can always handle symmetric errors simply by returning
    # them as if asymmetric. However, leave this function here so that
    # it can be easily changed if we want.
    header = "READ TERR 1"
    if "UpperLimit" not in data[curve].columns:
        header = header + " 2"
    header = header + "\n"
    cols = data[curve].columns.tolist()
    cols = [x for x in cols if x != "ObsID"]
    header = header + "!" + sep.join(cols)
    return header


def _saveDFToDisk(data, fname, cols, header, sep, qdpH, asQDP, clobber=False, silent=False, verbose=True):
    """Actually save a DataFrame to disk.

    Parameters
    ----------

    data : DataFrame
        The DataFrame to save.

    fname : str
        The file name.

    cols : list
        The columns to write.

    header : bool
        Whether to write a header columns.

    sep : str
        The column delimiter.

    qdpH : str or None
        The qdp-style header (READ lines).

    asQDP : bool
        Whether to save as a qdp file.

    clobber : bool, optional
        Whether to overwrite files if they exist (default: ``False``).

    silent : bool, optional
        Whether to suppress all output (default: ``True``).

    verbose : bool, optional
        Whether to write verbose output (default: ``False``).

    Returns
    -------

    bool
        Whether everything worked OK or not.

    """
    if verbose:
        silent = False

    if os.path.exists(fname) and not clobber:
        if not silent:
            print(f"Cannot write `{fname}`, already exists and clobber=False.")
            return False

    # print("COLUMNS: ",data.columns.tolist())
    # print("LIST: ", cols)

    txtBuffer = io.StringIO()
    data.to_csv(txtBuffer, index=False, sep=sep, header=False, columns=cols)
    with open(fname, "w") as outfile:
        if verbose:
            print(f"Writing file: `{fname}`")
        if asQDP and qdpH is not None:
            outfile.write(f"{qdpH}\n")
        if header:
            if asQDP:
                outfile.write("!")
            outfile.write(sep.join(cols) + "\n")

        outfile.write(txtBuffer.getvalue())

    return True


# --------------------------------------------------------------------
# Product-related functions, that are shared between modules
# --------------------------------------------------------------------

# -----------------------------
# Light curve related functions


def _getLightCurve(
    type,
    objectID,
    returnData=False,
    saveData=True,
    nosys="no",
    incbad="yes",
    destDir="lc",
    prefix="",
    clobber=False,
    skipErrors=False,
    silent=True,
    verbose=False,
):
    """Get a specific XRT light curve.

    This function is not intended to be called directly by the user,
    but rather, via wrappers which will work out 'type' and 'objectID'.

    This function can do two things:

    * Download light curve data and return it as a data structure.

    * Save the light curve files to disk.

    These are not mutually exclusive; you can do them both.

    Parameters
    ----------

    type : str
        The type of light curve to get (e.g. 'GRB', 'REBIN_LC' etc.)

    objectID : int
        The ObjectID

    returnData : bool, optional
        Whether to return the organised light curve data from this
        function (default: ``False``).

    saveData : bool, optional
        Whether to save the light curve data to disk (default:
        ``False``).

    nosys : str
        Whether to return the light curves from which the WT
        systematic error has been removed. Can be 'yes', 'no',
        'both' -- the latter returning datasets with and without the
        systematic error (default: 'no').

    incbad : str
        Whether to return the light curves in which the datapoints
        flagged as potentially unreliable to due centroiding issues
        have been included. Can be 'yes', 'no', 'both' -- the latter
        returning datasets with and without the unreliable
        datapoints (default: 'yes').

    destDir : str, optional
        The directory in which to save the light curves (default: "lc").

    prefix : str, optional
        A prefix to prepend to the file names before saving.

    clobber : bool, optional
        Whether to overwrite files that already exist
        (default: ``False``).

    skipErrors : bool, optional
        Whether to continue if a problem occurs with one file
        (default: ``False``).

    silent : bool, optional
        Whether to suppress all output (default: ``True``).

    verbose : bool, optional
        Whether to write verbose output (default: ``False``).

    Returns
    -------
    dict
        A dictionary with the light curve information as described
        above.
    """
    if verbose:
        silent = False

    if not (returnData or saveData):
        raise RuntimeError("I have to do at least one of save or return!")

    if isinstance(incbad, bool):
        if incbad:
            incbad = "yes"
        else:
            incbad = "no"

    if isinstance(nosys, bool):
        if nosys:
            nosys = "yes"
        else:
            nosys = "no"

    sendData = {"type": type, "objectID": objectID, "incbad": incbad, "nosys": nosys}

    tmp = base.submitAPICall("downloadLightCurve", sendData, verbose=verbose, minKeys=("Datasets",))
    ret = None

    if "NOLC" in tmp:
        if returnData:
            ret = tmp
    else:
        if returnData:
            ret = _handleLightCurve(tmp, silent=silent, verbose=verbose)
        else:
            ret = {}
            # If we didn't want to return, let's not waste CPU time making the
            # DataFrames, but we still need the URLs:
            url = {}
            for key in tmp["Datasets"]:
                tmpKey = f"{key}Data"
                if tmpKey not in tmp:
                    raise RuntimeError(f"Expected to find `{key}` data, but it is not present.")

                if "URL" in tmp[tmpKey]:
                    url[key] = tmp[tmpKey]["URL"]

            if len(url) > 0:
                ret["URLs"] = url
                ret["Datasets"] = tmp["Datasets"]

        if saveData:
            _saveLightCurveFromURL(
                ret,
                destDir=destDir,
                clobber=clobber,
                incbad=incbad,
                nosys=nosys,
                prefix=prefix,
                skipErrors=skipErrors,
                silent=silent,
                verbose=verbose,
            )

    if returnData:
        return ret


def _saveLightCurveFromDict(
    data,
    destDir="lc",
    asQDP=False,
    whichDatasets="all",
    clobber=False,
    header=False,
    sep=",",
    prefix="",
    suff=None,
    timeFormatInFname=False,
    binningInFname=False,
    skipErrors=False,
    silent=False,
    verbose=False,
    **kwargs,
):
    """Save a light curve dict as files on disk.

    This function take a standard light curve ``dict`` and saves the
    requested datasets to disk following the conventions set by the
    function parameters.

    Parameters:
    -----------

    data : dict
        A standard light curve ``dict`` of the data to save.

    destDir : str, optional
        The directory in which to save the files (default: 'lc').

    asQDP : bool, optional
        Whether to save in qdp format. Overrides ``sep`` and ``suff``
        (default: ``False``).

    whichDatasets : list or str, optional
        A list of the keys identifying the datasets to save, or 'all'
        (default: 'all').

    clobber : bool
        Whether to overwrite files if they exist.

    header : bool, optional
        Whether to print a header row (default: ``False``).

    sep : str, optional
        Separator to use for columns in the file (default: ',').

    prefix : str, optiona;
        An optional string to prepend to the dataset label in creating
        the file names.

    suff : str, optional
        The file suffix to use. If ``None`` then it will be "qdp" if
        ``asQDP`` is ``True``, else ".dat" (default: ``None``).

    timeFormatInFname : bool, optional
        Whether the filename should include the time format of the light
        curve (default: ``False``).

    binningInFname : bool, optional
        Whether the filename should include the binning method of the
        light curve (default: ``False``).

    skipErrors : bool, optional
        If an error occurs saving a file, do not raise a RuntimeError
        but simply continue to the next file (default ``False``).

    silent : bool, optional
        Whether to suppress all output (default: ``True``).

    verbose : bool, optional
        Whether to write verbose output (default: ``False``).

    **kwargs: dict, optional
        This exists so that if extra parameters are passed to this
        function they do not cause errors; this is necessary because
        the data.SXPS.getLightCurves() module will pass some extra
        arguments to this function.

    """
    if verbose:
        silent = False

    if "Datasets" not in data:
        raise ValueError("The `data` parameter should be a light curve dict. It isn't.")

    theseCurves = []
    if (whichDatasets is None) or (whichDatasets == "all"):
        theseCurves = data["Datasets"]
    else:
        theseCurves = whichDatasets

    base._createDir(destDir, silent=silent, verbose=verbose)

    # Grab some helper vars
    b = data["Binning"]
    t = data["TimeFormat"]
    if t == "Swift MET":
        t = "MET"

    if suff is None:
        if asQDP:
            suff = "qdp"
        else:
            suff = "dat"

    for c in theseCurves:
        fname = f"{destDir}/{prefix}{c}"
        if timeFormatInFname:
            fname = fname + f"_{t}"
        if binningInFname:
            fname = fname + f"_{b}"
        fname = fname + f".{suff}"

        if c not in data["Datasets"]:
            if not silent:
                print(f"Not saving {prefix}{c} as this curve does not exist.")
            continue

        cols = data[c].columns.tolist()
        qdpheader = None
        if asQDP:
            cols = [x for x in cols if x != "ObsID"]
            qdpheader = _getQDPHeader(data, c, sep)

        ok = _saveDFToDisk(
            data[c],
            fname,
            cols,
            header,
            sep,
            qdpheader,
            asQDP,
            clobber=clobber,
            silent=silent,
            verbose=verbose,
        )
        if not (ok or skipErrors):
            raise RuntimeError("Cannot write `{fname}`")


# ---------
# Rebin fns


def _rebinLightCurve(type, objectID, binMeth=None, silent=True, verbose=False, **kwargs):
    """Submit a job to request a light curve.

     This function is not intended to be called directly by the user,
     but rather, via wrappers which will work out 'type' and 'objectID'.

     Parameters
     ----------

    type : str
         The type of light curve to get (e.g. 'GRB', 'REBIN_LC' etc.)

     objectID : int
         The ObjectID

     binMeth : str
         What binning method to apply ('snapshot', 'observation', 'time',
         'counts').

     silent : bool, optional
         Whether to suppress all output (default: ``True``).

     verbose : bool, optional
         Whether to write verbose output (default: ``False``).

     kwargs : dict
         Parameters defining the binning; please see the online
         documentation for a full description of these.

     Returns
     -------

     int
         The job identifier, which is needed to retrieve your data.

    """
    if verbose:
        silent = False  # noqa

    sendData = {"type": type, "objectID": objectID, "binMeth": binMeth}
    sendData.update(kwargs)

    tmp = base.submitAPICall("rebinLightCurve", sendData, verbose=verbose, minKeys=("JobID",))
    t = int(tmp["JobID"])
    if verbose:
        print(f"Success: JobID={t}")
    return t


def _checkRebinStatus(JobID, silent=True, verbose=False):
    """Check the status of a rebin job.

    This takes the JobID (returned by ``rebinLightCurve()``) and returns
    its status.

    Parameters
    ----------

    JobID : int
        The Job ID.

    silent : bool, optional
        Whether to suppress all output (default: ``True``).

    verbose : bool, optional
        Whether to write verbose output (default: ``False``).

    Returns
    -------

    dict
        A dictionary with keys:

        statusCode : int
            :A code indicating the status of the job.

        statusText : str
            :A textual description of the job status.

    """
    if verbose:
        silent = False  # noqa

    if not isinstance(JobID, int):
        raise ValueError("JobID must be an int")

    sendData = {"JobID": JobID}

    tmp = base.submitAPICall("checkRebinStatus", sendData, verbose=verbose, minKeys=("status", "text"))

    ret = {"statusCode": tmp["status"], "statusText": tmp["text"]}

    return ret


def _rebinComplete(JobID):
    """Check whether a rebin job is complete

    Parameters
    ----------

    JobID : int
        The Job ID.

    Returns
    -------

    bool
        Whether the job is complete or not.

    """
    if not isinstance(JobID, int):
        raise ValueError("JobID must be an int")

    tmp = _checkRebinStatus(JobID, silent=True, verbose=False)

    if tmp["statusCode"] == 4:
        return True
    else:
        return False


def _cancelRebin(JobID, silent=True, verbose=False):
    """Cancels a rebin job.

    Parameters
    ----------

    JobID : int
        The Job ID.

    silent : bool, optional
        Whether to suppress all output (default: ``True``).

    verbose : bool, optional
        Whether to write verbose output (default: ``False``).

    Returns
    -------

    bool
        Whether the was cancelled or not.

    """
    if verbose:
        silent = False

    if not isinstance(JobID, int):
        raise ValueError("JobID must be an int")

    sendData = {"JobID": JobID}
    tmp = base.submitAPICall("cancelRebin", sendData, verbose=verbose, skipErrors=True)

    if "OK" in tmp:
        return True
    else:
        if verbose or not silent:
            print("Cannot cancel job - it may have already completed.")
        return False


# ---------------------------------
# Spectrum related functions


def _saveSpectrum(
    data,
    destDir="spec",
    spectra="all",
    specSubDirs=False,
    saveImages=True,
    saveData=True,
    prefix="",
    extract=False,
    removeTar=False,
    clobber=False,
    skipErrors=False,
    silent=True,
    verbose=False,
):
    """Internal function to save spectra to disk.

    This saves the spectra files, identified in the supplied dict,
    to disk. This expects to receive the standard dict structure
    returned by the API back end.

    Parameters
    ----------

    data : dict
        The object containing the data.

    destDir : str, optional
        The directory in which to save the spectra (default: "spec").

    spectra : str or list, optional
        Which spectra to save data for ("all" or a list; default "all").

    specSubDirs : bool, optional
        Whether to save each spectrum to its own subdirectory (with the
        spectrum name) (default: ``False``).

    saveImages : bool, optional
        Whether to save the gif images (default ``True``).

    saveData : bool, optional
        Whether to save the actual spectral data (default ``True``).

    prefix : str, optional
        A prefix to prepend to the image file names before saving.

    extract : bool, optional
        Whether to extract the spectral files from the
        tar archive (default ``False``).

    removeTar : bool, optional
        Whether to remove the tar file after extracting. **This
        parameter is ignored unless ``extract`` is ``True``**
        (default: ``False``).

    clobber : bool, optional
        Whether to overwrite files if they exist (default ``False``).

    skipErrors : bool, optional
        If an error occurs saving a file, do not raise a RuntimeError
        but simply continue to the next file (default ``False``).

    silent : bool, optional
        Whether to suppress all output (default: ``True``).

    verbose : bool, optional
        Whether to write verbose output (default: ``False``).

    """
    if verbose:
        silent = False

    # Structure of data is: regions => modes => models => spectra
    # Datafile appears by regions
    # image by spectrum

    # Step over regions
    if "rnames" not in data:
        raise ValueError("data dict is missing the 'Spectra' entry")

    if not (saveData or saveImages):
        if not silent:
            print("Nothing to do, not selected anything to save.")
            return

    base._createDir(destDir, silent=silent, verbose=verbose)

    for rname in data["rnames"]:
        if (spectra == "all") or (rname in spectra):
            if verbose:
                print(f"Saving `{rname}` spectrum")
            path = destDir
            if specSubDirs:
                path = f"{destDir}/{rname}"
                base._createDir(path, silent=silent, verbose=verbose)

            # The data file is at the spectrum level so if we wanted it, save it now
            if saveData and ("DataFile" in data[rname]):
                url = data[rname]["DataFile"]
                ok = _saveTar(
                    url,
                    path,
                    prefix=prefix,
                    extract=extract,
                    removeTar=removeTar,
                    clobber=clobber,
                    silent=silent,
                    verbose=verbose,
                )
                if not ok:
                    if not skipErrors:
                        raise RuntimeError(f"Cannot save/extract {url} in {path}/")

            if saveImages and ("Modes" in data[rname]):
                for mode in data[rname]["Modes"]:
                    for model in data[rname][mode]["Models"]:
                        if "Image" in data[rname][mode][model]:
                            url = data[rname][mode][model]["Image"]
                            ok = _saveURLToFile(
                                url,
                                path,
                                prefix=prefix,
                                clobber=clobber,
                                silent=silent,
                                verbose=verbose,
                            )
                        if not ok:
                            if skipErrors:
                                continue
                            else:
                                raise RuntimeError(f"Cannot save {url} into {path}/")


def _saveTar(
    url,
    path,
    prefix="",
    extract=False,
    removeTar=False,
    strip=False,
    clobber=False,
    silent=True,
    verbose=False,
    **kwargs,
):
    ok = _saveURLToFile(url, path, prefix=prefix, clobber=clobber, silent=silent, verbose=verbose)
    if not ok:
        return False

    if extract:
        fname = os.path.basename(url)
        fname = f"{path}/{fname}"

        os.chdir(path)
        if not silent:
            print(f"Extracting `{fname}`")
        comm = ["tar", "-xzf", fname]
        if strip:
            comm.insert(1, "--strip-components=1")
        if verbose:
            comm.insert(1, "-v")
            # pcomm = " ".join(comm)
            # print(f"Calling `{pcomm}`")

        status = subprocess.run(comm, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        if status.returncode != 0:
            if verbose:
                print(f"tar command exited with {status.returncode}")
            return False
        if verbose:
            print(status.stdout.decode("utf-8"))
            print(status.stderr.decode("utf-8"))
        if removeTar:
            os.unlink(fname)
            if verbose:
                print(f"Removing file {fname}")

    return True



from .api_common import TOOAPI_Baseclass
from .api_daterange import TOOAPI_Daterange
from .api_resolve import TOOAPI_AutoResolve
from .api_skycoord import TOOAPI_SkyCoord
from .api_status import TOOStatus
from .swift_clock import TOOAPI_ClockCorrect
from .swift_instruments import TOOAPI_Instruments
from .swift_obsid import TOOAPI_ObsID


class Swift_CalendarEntry(
    TOOAPI_Baseclass,
    TOOAPI_Instruments,
    TOOAPI_ClockCorrect,
    TOOAPI_Daterange,
    TOOAPI_SkyCoord,
):
    """Class for a single entry in the Swift TOO calendar.

    Attributes
    ----------
    start : datetime
        start time of calendar entry
    stop : datetime
        stop time of calendar entry
    xrt_mode : str
        XRT mode of calendar entry
    uvot_mode : str
        UVOT mode of calendar entry
    bat_mode : str
        BAT mode of calendar entry
    duration : int
        exposure time of calendar entry in seconds
    asflown: float
        estimated exposure time in seconds
    merit: float
        figure of merit of calendar entry
    targetid : int
        target ID  of the observation
    ra : float
        Right Ascension of pointing in J2000 (decimal degrees)
    dec : float
        Declination of pointing in J2000 (decimal degrees)
    """

    # Set up Core API values
    _attributes = [
        "start",
        "stop",
        "xrt_mode",
        "bat_mode",
        "uvot_mode",
        "duration",
        "asflown",
        "merit",
        "targetid",
        "ra",
        "dec",
    ]
    # Variable names
    _varnames = {
        "start": "Start",
        "stop": "Stop",
        "xrt_mode": "XRT Mode",
        "bat_mode": "BAT Mode",
        "uvot_mode": "UVOT Mode",
        "duration": "Exposure (s)",
        "asflown": "AFST (s)",
        "merit": "Merit",
        "ra": "Right Ascension (deg)",
        "dec": "Declination (deg)",
        "targetid": "Target ID",
    }
    api_name = "Swift_Calendar_Entry"

    def __init__(self):
        # Parameters
        # Start and end times of the observing window. Use datetime for coordinated window
        self.start = None
        # and timedelta for relative offsets, where the start time can be
        # variable, but the monitoring cadence isn't even.
        self.stop = None
        self.merit = None
        self.targetid = None
        self.duration = None  # Exposure time in seconds
        self.asflown = None  # Amount of exposure taken
        self.ignorekeys = True

    def __getitem__(self, key):
        if key in self._parameters:
            return getattr(self, key)

    @property
    def _table(self):
        parameters = ["start", "stop", "xrt_mode", "uvot_mode", "duration", "asflown"]
        header = [self._varnames[row] for row in parameters]
        return header, [[getattr(self, row) for row in parameters]]


class Swift_Calendar(
    TOOAPI_Baseclass,
    TOOAPI_ClockCorrect,
    TOOAPI_Daterange,
    TOOAPI_SkyCoord,
    TOOAPI_AutoResolve,
    TOOAPI_ObsID,
):
    """Class that fetches entries in the Swift Planning Calendar, which
    are scheduled as part of a TOO request.

    Attributes
    ----------
    begin : datetime
        begin time of visibility window
    end : datetime
        end time of visibility window
    length : timedelta / int
        length of visibility window
    ra : float
        Right Ascension of target in J2000 (decimal degrees)
    dec : float
        Declination of target in J2000 (decimal degrees)
    radius : float
        Search radius in degrees
    skycoord : SkyCoord
        SkyCoord version of RA/Dec if astropy is installed
    username : str
        username for TOO API (default 'anonymous')
    shared_secret : str
        shared secret for TOO API (default 'anonymous')
    too_id : int
        Unique TOO identifying number
    entries : list
        list of calendar entries returned by query (`Swift_CalendarEntries`)
    status : Swift_TOOStatus
        Status of API request
    """

    # Core API definitions
    api_name = "Swift_Calendar"
    _parameters = [
        "username",
        "too_id",
        "begin",
        "end",
        "ra",
        "dec",
        "begin",
        "end",
        "radius",
        "targetid",
    ]
    _attributes = ["status", "entries"]
    # Local parameters
    _local = ["shared_secret", "length", "name"]
    # Subclasses used by class
    _subclasses = [Swift_CalendarEntry, TOOStatus]

    def __init__(self, *args, **kwargs):
        """
        Parameters
        ----------
        username : str
            username for TOO API (default 'anonymous')
        shared_secret : str
            shared secret for TOO API (default 'anonymous')
        too_id : int
            Unique TOO identifying number
        """
        # Parameters
        self.entries = list()
        self.username = "anonymous"
        self.too_id = None
        self.status = TOOStatus()
        self.length = None
        self.ra = None
        self.dec = None
        self.targetid = None

        # Read in arguements
        self._parseargs(*args, **kwargs)

        # See if we pass validation from the constructor, but don't record
        # errors if we don't
        if self.validate():
            self.submit()
        else:
            self.status.clear()

    def __getitem__(self, number):
        return self.entries[number]

    def __len__(self):
        return len(self.entries)

    def validate(self):
        if (
            self.too_id is not None
            or self.length is not None
            or (self.ra is not None and self.dec is not None)
            or self.targetid is not None
        ):
            return True
        else:
            self.status.error("TOO ID is not set.")
            return False

    @property
    def _table(self):
        """Table of Calendar details"""
        table = list()
        for i in range(len(self.entries)):
            table.append([i] + self.entries[i]._table[-1][0])
        if len(self.entries) > 0:
            header = ["#"] + self.entries[0]._table[0]
        else:
            header = []
        return header, table


# Shorthand alias
Calendar = Swift_Calendar
CalendarEntry = Swift_CalendarEntry
# Back compat
Swift_Calendar_Entry = Swift_CalendarEntry



import json
import re
import textwrap
import warnings
from datetime import date, datetime, timedelta, timezone
from time import sleep

import requests
from dateutil import parser
import jwt
from tabulate import tabulate

from .version import version_tuple

# Make Warnings a little less weird
formatwarning_orig = warnings.formatwarning
warnings.formatwarning = lambda message, category, filename, lineno, line=None: formatwarning_orig(
    message, category, filename, lineno, line=""
)

# Define the API version
api_version = f"{version_tuple[0]}.{version_tuple[1]}"

# Next imports are not dependancies, but if you have them installed, we'll use
# them
try:
    import keyring

    # Check that keyring actually is set up and working
    if keyring.get_keyring().name != "fail Keyring":
        keyring_support = True
    else:
        keyring_support = False
except ImportError:
    keyring_support = False

# Check if we have astropy support
HAS_ASTROPY = False
try:
    import astropy.units as u
    from astropy.time import Time

    HAS_ASTROPY = True
except ImportError:
    pass


# Convert degrees to radians
dtor = 0.017453292519943295

# Regex for matching date, time and datetime strings
_date_regex = r"^[0-2]\d{3}-(0?[1-9]|1[012])-([0][1-9]|[1-2][0-9]|3[0-1])?$"
_time_regex = r"^([0-9]:|[0-1][0-9]:|2[0-3]:)[0-5][0-9]:[0-5][0-9]+(\.\d+)?$"
_iso8601_regex = r"^([\+-]?\d{4}(?!\d{2}\b))((-?)((0[1-9]|1[0-2])(\3([12]\d|0[1-9]|3[01]))?|W([0-4]\d|5[0-2])(-?[1-7])?|(00[1-9]|0[1-9]\d|[12]\d{2}|3([0-5]\d|6[1-6])))([T\s]((([01]\d|2[0-3])((:?)[0-5]\d)?|24\:?00)([\.,]\d+(?!:))?)?(\17[0-5]\d([\.,]\d+)?)?([zZ]|([\+-])([01]\d|2[0-3]):?([0-5]\d)?)?)?)?$"
_datetime_regex = r"^[0-2]\d{3}-(0?[1-9]|1[012])-([0][1-9]|[1-2][0-9]|3[0-1]) ([0-9]:|[0-1][0-9]:|2[0-3]:)[0-5][0-9]:[0-5][0-9]+(\.\d+)?$"
_float_regex = r"^[+-]?(?=\d*[.eE])(?=\.?\d)\d*\.?\d*(?:[eE][+-]?\d+)?$"
_int_regex = r"^(0|[1-9][0-9]+)$"

# Submission URL
API_URL = "https://www.swift.psu.edu/toop/submit_json.php"


def convert_to_dt(value, isutc=False, outfunc=datetime):
    """Convert various date formats to swiftdatetime or datetime

    Parameters
    ----------
    value : varies
        Value to be converted.
    isutc : bool, optional
        Is the value in UTC, by default False
    outfunc : datetime / swiftdatetime, optional
        What format should the output be? By default datetime, can be
        swiftdatetime

    Returns
    -------
    datetime / swiftdatetime
        Returned datetime / swiftdatetime object

    Raises
    ------
    TypeError
        Raised if incorrect format is given for conversion.
    """

    if type(value) == str:
        if re.match(_datetime_regex, value):
            if "." in value:
                # Do this because "fromisoformat" is restricted to 0, 3 or 6 decimal plaaces
                dtvalue = outfunc.strptime(value, "%Y-%m-%d %H:%M:%S.%f")
            else:
                dtvalue = outfunc.strptime(value, "%Y-%m-%d %H:%M:%S")
        elif re.match(_date_regex, value):
            dtvalue = outfunc.strptime(f"{value} 00:00:00", "%Y-%m-%d %H:%M:%S")
        elif re.match(_iso8601_regex, value):
            dtvalue = parser.parse(value)
            if dtvalue.tzinfo is None:
                warnings.warn(
                    "ISO8601 formatted dates should be supplied with timezone. ISO8601 dates with no timezone will be assumed to be localtime and then converted to UTC."
                )
            dtvalue = dtvalue.astimezone(timezone.utc).replace(tzinfo=None)
        else:
            raise ValueError("Date/time given as string should 'YYYY-MM-DD HH:MM:SS' or ISO8601 format.")
    elif type(value) == date:
        dtvalue = outfunc.strptime(f"{value} 00:00:00", "%Y-%m-%d %H:%M:%S")
    elif type(value) == outfunc:
        if value.tzinfo is not None:
            # Strip out timezone info and convert to UTC
            value = value.astimezone(timezone.utc).replace(tzinfo=None)
        dtvalue = value  # Just pass through un molested
    elif type(value) == swiftdatetime and outfunc == datetime or type(value) == datetime and outfunc == swiftdatetime:
        if type(value) == datetime and value.tzinfo is not None:
            # Strip out timezone info and convert to UTC
            value = value.astimezone(timezone.utc).replace(tzinfo=None)
        dtvalue = outfunc.fromtimestamp(value.timestamp())
    elif value is None:
        dtvalue = None
    elif HAS_ASTROPY and type(value) is Time:
        dtvalue = value.datetime
    else:
        raise TypeError(
            'Date should be given as a datetime, astropy Time, or as string of format "YYYY-MM-DD HH:MM:SS"'
        )
    if outfunc is swiftdatetime and dtvalue.isutc != isutc:
        dtvalue.isutc = isutc
    return dtvalue


def _tablefy(table, header=None):
    """Simple HTML table generator

    Parameters
    ----------
    table : list
        Data for table
    header : list
        Headers for table, by default None

    Returns
    -------
    str
        HTML formatted table.
    """

    tab = "<table>"
    if header is not None:
        tab += "<thead>"
        tab += "".join([f"<th style='text-align: left;'>{head}</th>" for head in header])
        tab += "</thead>"

    for row in table:
        tab += "<tr>"
        # Replace any carriage returns with <br>
        row = [f"{col}".replace("\n", "<br>") for col in row]
        tab += "".join([f"<td style='text-align: left;'>{col}</td>" for col in row])
        tab += "</tr>"
    tab += "</table>"
    return tab


class TOOAPI_Baseclass:
    """Mixin for TOO API Classes. Most of these are to do with reading and
    writing classes out as JSON/dicts."""

    # Set api_version for all classes
    api_version = api_version
    # Ignore any keys you don't understand
    ignorekeys = False
    username = "anonymous"
    _shared_secret = None
    # Submission timeout
    timeout = 120  # 2 mins
    # By default all API dates are in Swift Time
    _isutc = False

    # Parameters that are _local
    _local = []
    # Parameters that get sent in API submission
    _parameters = []
    # Attributes that get send back by API
    _attributes = []

    @property
    def shared_secret(self):
        if self._shared_secret is None and self.username != "anonymous":
            # Try to fetch password using keyring, if available
            if keyring_support:
                self._shared_secret = keyring.get_password("swifttools.swift_too", self.username)
            else:
                raise Exception("Warning: keyring support not available. Please set shared_secret manually.")
        elif self.username == "anonymous":
            return "anonymous"
        return self._shared_secret

    @shared_secret.setter
    def shared_secret(self, secret):
        if self.username != "anonymous" and (self.username is not None or self.username == ""):
            # Try to remember the password using keyring if available
            if keyring_support:
                try:
                    keyring.set_password("swifttools.swift_too", self.username, secret)
                except keyring.errors.PasswordSetError:
                    # This error is raised in if keyring support is enabled on
                    # macOS, but the person isn't running the code interactively
                    # or if they're logged in via ssh. Just trap and move on to
                    # avoid code crashing.
                    pass

        if self.username != "anonymous":
            self._shared_secret = secret

    @property
    def _table(self):
        """Table of details of the class"""
        _parameters = self._parameters + self._attributes
        header = ["Parameter", "Value"]
        table = []
        for row in _parameters:
            value = getattr(self, row)
            if value is not None and value != [] and value != "":
                if row == "status" and type(value) != str:
                    table.append([row, value.status])
                elif type(value) == list:
                    table.append([row, "\n".join([f"{le}" for le in value])])
                else:
                    table.append([row, "\n".join(textwrap.wrap(f"{value}"))])
        return header, table

    def _repr_html_(self):
        if (
            hasattr(self, "status")
            and self.status == "Rejected"
            and self.status.__class__.__name__ == "Swift_TOO_Status"
        ):
            return "<b>Rejected with the following error(s): </b>" + " ".join(self.status.errors)
        else:
            header, table = self._table
            if len(table) > 0:
                return _tablefy(table, header)
            else:
                return "No data"

    def __str__(self):
        if (
            hasattr(self, "status")
            and self.status == "Rejected"
            and self.status.__class__.__name__ == "Swift_TOO_Status"
        ):
            return "Rejected with the following error(s): " + " ".join(self.status.errors)
        else:
            header, table = self._table
            if len(table) > 0:
                return tabulate(table, header, tablefmt="pretty", stralign="right")
            else:
                return "No data"

    def __repr__(self):
        name = self.__class__.__name__
        args = ",".join(
            [
                f"{row}='{getattr(self, row)}'"
                for row in self._parameters
                if getattr(self, row) is not None and getattr(self, row) != []
            ]
        )
        return f"{name}({args})"

    def _parseargs(self, *args, **kwargs):
        """Parse arguements given in __init__ to API classes"""
        # Parse arguments. Assume they're the same order as `_parameters` except not `username`
        for i in range(len(args)):
            setattr(self, self._parameters[i + 1], args[i])
        # Parse argument keywords
        for key in kwargs.keys():
            if key in self._parameters + self._local:
                setattr(self, key, kwargs[key])
            else:
                raise TypeError(f"{self.api_name} got an unexpected keyword argument '{key}'")

    @property
    def too_api_dict(self):
        """Dictionary version of TOO API object"""
        too_api_dict = dict()
        too_api_dict["api_name"] = self.api_name
        too_api_dict["api_version"] = self.api_version
        too_api_dict["api_data"] = self.api_data
        return too_api_dict

    @property
    def api_data(self):
        """Convert class parameters and data into a dictionary"""
        data = dict()
        for param in self._parameters:
            value = getattr(self, param)
            if value is not None:
                # Note just convert swiftdatetime into isoformat strings for now
                if "api_data" in dir(value) and value.api_name != "swiftdatetime":
                    data[param] = value.too_api_dict
                elif type(value) == list or type(value) == tuple:

                    def conv(x):
                        return f"{x}" if not hasattr(x, "too_api_dict") else x.too_api_dict

                    data[param] = [conv(entry) for entry in value]
                elif (
                    type(value) == datetime
                    or type(value) == date
                    or type(value) == timedelta
                    or type(value) == swiftdatetime
                ):
                    data[param] = f"{value}"
                # Detect and convert astropy Time
                elif HAS_ASTROPY and type(value) == Time:
                    data[param] = f"{value.datetime}"
                elif HAS_ASTROPY and type(value) == u.quantity.Quantity:
                    # For any astropy units not handled explicitly, pass as string
                    data[param] = f"{value}"
                else:
                    data[param] = value
        return data

    @property
    def jwt(self):
        """JWT version of TOO API Object"""
        return jwt.encode(self.too_api_dict, self.shared_secret, algorithm="HS256")

    def __convert_dict_entry(self, entry):
        """Parse data entry from a dictionary (usually originating as a JSON) to
        convert into Python data types. Danger! Danger Will Robinson! Recursion!
        Recursion!"""
        # Parse a JSON entry
        if type(entry) == dict and "api_name" in entry.keys():
            index = [s.api_name for s in self._subclasses].index(entry["api_name"])
            if self._subclasses[index] == swiftdatetime:
                # Handle swiftdatetime as it cannot be created with no arguments
                val = swiftdatetime.frommet(
                    entry["api_data"]["met"],
                    utcf=entry["api_data"]["utcf"],
                    isutc=entry["api_data"]["isutc"],
                )
            else:
                # Handle everything else
                val = self._subclasses[index]()
                val.__read_dict(entry["api_data"])
        # Parse a list of items
        elif type(entry) == list:
            # Parse a list of jsons
            val = list()
            for subvalue in entry:
                # Hey, we must have some handy function for parsing values, right?
                val.append(self.__convert_dict_entry(subvalue))
        # Parse all other values
        else:
            val = False
            if entry:
                # Check if these are dates, datetimes or times by regex matching
                match = re.match(_time_regex, str(entry))
                if match is not None:
                    hours, mins, secs = match[0].split(":")
                    hours = int(hours)
                    mins = int(mins)
                    secs = int(float(secs))
                    millisecs = int(1000.0 * secs % 1)
                    val = timedelta(hours=hours, minutes=mins, seconds=secs, milliseconds=millisecs)

                # Parse dates into a datetime.date
                match = re.match(_date_regex, str(entry))
                if match is not None:
                    val = datetime.strptime(match[0], "%Y-%m-%d").date()

                # Parse a date/time into a datetime.datetime
                match = re.match(_datetime_regex, str(entry))
                if match is not None:
                    val = convert_to_dt(match[0], self._isutc)

                # If it's a float, convert it
                if type(entry) == str:
                    match = re.match(_float_regex, entry)
                    if match is not None:
                        val = float(entry)

                # If it's an int, convert it
                if type(entry) == str:
                    match = re.match(_int_regex, entry)
                    if match is not None:
                        val = int(entry)

            # None of the above? It is what it is.
            if val is False:
                val = entry

        return val

    def __set_status(self, newstatus):
        if hasattr(self, "status"):
            if type(self.status) == str:
                self.status = newstatus
            else:
                self.status.status = newstatus
        else:
            print(f"STATUS: {newstatus}")

    def __set_error(self, newerror):
        if hasattr(self, "status"):
            if type(self.status) == str:
                self.error(newerror)
            else:
                self.status.error(newerror)
        else:
            print(f"ERROR: {newerror}")

    def __read_dict(self, data_dict):
        """Read from a dictionary values for the class"""
        for key in data_dict.keys():
            if key in self._parameters or key in self._attributes:
                val = self.__convert_dict_entry(data_dict[key])
                if val is not None:  # If value is set to None, then don't change the value
                    if hasattr(self, f"_{key}"):
                        setattr(self, f"_{key}", val)
                    else:
                        setattr(self, key, val)
            else:
                if not self.ignorekeys:  # If keys exist in JSON we don't understand, fail out
                    self.__set_error(f"Unknown key in JSON file: {key}")
                    return False
        return True  # No errors

    @property
    def submit_url(self):
        """Generate a URL that submits the TOO API request"""
        url = f"{API_URL}?jwt={self.jwt}"
        return url

    def queue(self, post=True):
        """Validate and submit a TOO API job to the queue for processing."""
        # Make sure a shared secret is set
        if self.shared_secret is None:
            self.__set_error("shared_secret not set, cannot submit job.")
            return False
        # Make sure it passes validation checks
        if not self.validate():
            self.__set_error("Swift TOO API submission did not pass internal validation checks.")
            return False

        return self.__submit_jwt(post=post)

    @property
    def complete(self, post=True):
        """Check if a queued job is completed"""
        # Send request to TOO API server
        self.__submit_jwt(post=post)
        # Check the status of the request
        if self.status != "Queued" and self.status != "Processing":
            return True
        return False

    def __submit_jwt(self, post=True):
        """Submit JWT request to the TOO API server, read in the reply and parse it."""
        # Don't submit an accepted or rejected request more than once
        if self.status == "Accepted":
            return True
        elif self.status == "Rejected":
            return False

        # Which way will we fetch the data?
        if post:
            r = self.__submit_post()
        else:
            r = self.__submit_get()

        # See how sucessful we were
        if r.status_code == 200:
            # True to decode the returned JSON. Return an error if it doesn't work.
            try:
                too_api_dict = json.loads(r.text)
            except json.decoder.JSONDecodeError:
                self.__set_error("Failed to decode JSON. Please check that your shared secret is correct.")
                self.__set_status("Rejected")
                return False

            # Determine that we are running the correct API version
            if too_api_dict["api_version"] != self.api_version:
                self.__set_error(
                    f"API version mismatch. Remote version {too_api_dict['api_version']} vs"
                    f"local version {self.api_version}. Ensure you're running the latest API code."
                )
                self.__set_status("Rejected")
                return False

            # Determine if the returned JSON is the full result or just a status message
            if too_api_dict["api_name"] == self.api_name:
                self.__read_dict(too_api_dict["api_data"])
            elif too_api_dict["api_name"] == "Swift_TOO_Status":
                self.status.__read_dict(too_api_dict["api_data"])
        else:
            self.__set_error(f"HTTP Submit failed with error code {r.status_code}")
            self.__set_status("Rejected")
            return False
        return True

    def submit(self, timeout=None, post=True):
        """Queue up a TOO API job, then wait for it to complete. Default behaviour is to keep
        checking if the submission has been processed by the TOO_API server every 1 seconds
        until the timeout (default 120s) has been reached."""
        # Update timeout value from default if passed
        if timeout is not None:
            self.timeout = timeout

        # Submit the job to the queue
        ustart = datetime.now().timestamp()
        if not self.queue(post=post):
            self.__set_error("Failed to queue job.")
            self.__set_status("Rejected")
            return False

        # Check if the Queued job is complete for up to *timeout* seconds (120s default)
        while datetime.now().timestamp() - ustart < self.timeout and not self.complete:
            sleep(1)

        # If the job is still Queued, report that it has timed out as an error
        if self.status == "Queued" or self.status == "Processing":
            self.__set_error("Queued job timed out.")
            return False
        # Or else return True or False if it has been Accepted or Rejected
        else:
            if self.status == "Accepted":
                # Run post processing script to do things
                self._post_process()
                return True
            else:
                return False

    def _post_process(self):
        """Placeholder method. Things to do after values are returned from API."""
        pass

    def __submit_get(self):
        """Submit the request through the web based API, as a JWT through GET (essentially a URL)"""
        return requests.get(self.submit_url)

    def __submit_post(self):
        """Submit the request through the web based API, as a JWT through POST (recommended)"""
        return requests.post(url=API_URL, verify=True, data={"jwt": self.jwt})


class swiftdatetime(datetime, TOOAPI_Baseclass):
    """Extend datetime to store met, utcf and swifttime. Default value is UTC"""

    api_name = "swiftdatetime"
    _parameters = ["met", "utcf", "swifttime", "utctime", "isutc"]

    def __new__(self, *args, **kwargs):
        return super(swiftdatetime, self).__new__(self, *args)

    def __init__(self, *args, **kwargs):
        self._met = None
        self.utcf = None
        self._swifttime = None
        self._utctime = None
        self._isutc = False
        self._isutc_set = False
        if "tzinfo" in kwargs.keys():
            raise TypeError("swiftdatetime does not support timezone information.")
        for key in kwargs.keys():
            setattr(self, key, kwargs[key])

    def __repr__(self):
        return f"swiftdatetime({self.year}, {self.month}, {self.day}, {self.hour}, {self.minute}, {self.second}, {self.microsecond}, isutc={self.isutc}, utcf={self.utcf})"

    def __sub__(self, other):
        """Redefined __sub__ to handle mismatched time bases"""
        if isinstance(other, swiftdatetime):
            if self.isutc != other.isutc and (self.utctime is None or other.utctime is None):
                raise ArithmeticError(
                    "Cannot subtract mismatched time zones with no UTCF"
                )  # FIXME - correct exception?

            if self.isutc is True and other.isutc is True or self.isutc is False and other.isutc is False:
                return super().__sub__(other)
            else:
                if self.isutc:
                    return super().__sub__(other.utctime)
                else:
                    return self.utctime.__sub__(other.utctime)
        else:
            value = super().__sub__(other)
            if hasattr(value, "isutc"):
                value.isutc = self.isutc
            return value

    def __add__(self, other):
        """Custom add for swiftdatetime. Note that UTCF is not preserved, on purpose."""
        value = super().__add__(other)
        if hasattr(value, "isutc"):
            value.isutc = self.isutc
        return value

    @property
    def isutc(self):
        return self._isutc

    @isutc.setter
    def isutc(self, utc):
        """Is this swiftdatetime based on UTC or Swift Time"""
        if self._isutc_set is not True:
            # If we change the time base for this, reset the values of swifttime and utctime
            self._isutc = utc
            self.swifttime = None
            self.utctime = None
            self._isutc_set = True
        else:
            raise AttributeError("Cannot set attribute isutc when previously set.")

    @property
    def met(self):
        if self.swifttime is not None:
            return (self.swifttime - datetime(2001, 1, 1)).total_seconds()

    @met.setter
    def met(self, met):
        self._met = met

    @property
    def utctime(self):
        if self._utctime is None:
            if self.isutc:
                self._utctime = datetime(
                    self.year,
                    self.month,
                    self.day,
                    self.hour,
                    self.minute,
                    self.second,
                    self.microsecond,
                )
            elif self.utcf is not None:
                self._utctime = self.swifttime + timedelta(seconds=self.utcf)
        return self._utctime

    @utctime.setter
    def utctime(self, utc):
        if isinstance(utc, datetime):
            # Ensure that utctime set to a pure datetime
            if utc.tzinfo is not None:
                utc = utc.astimezone(timezone.utc).replace(tzinfo=None)
        else:
            self._utctime = utc

    @property
    def swifttime(self):
        if self._swifttime is None:
            if not self.isutc:
                self._swifttime = datetime(
                    self.year,
                    self.month,
                    self.day,
                    self.hour,
                    self.minute,
                    self.second,
                    self.microsecond,
                )
            elif self.utcf is not None:
                self._swifttime = self.utctime - timedelta(seconds=self.utcf)
        return self._swifttime

    @swifttime.setter
    def swifttime(self, st):
        if isinstance(st, datetime):
            # Ensure that swifttime set to a pure datetime
            if st.tzinfo is not None:
                st = st.astimezone(timezone.utc).replace(tzinfo=None)
            self._swifttime = st
        else:
            self._swifttime = convert_to_dt(st)

    @property
    def _table(self):
        if self._isutc:
            header = ["MET (s)", "Swift Time", "UTC Time (default)", "UTCF (s)"]
        else:
            header = ["MET (s)", "Swift Time (default)", "UTC Time", "UTCF (s)"]
        return header, [[self.met, self.swifttime, self.utctime, self.utcf]]

    @classmethod
    def frommet(cls, met, utcf=None, isutc=False):
        """Construct a swiftdatetime from a given MET and (optional) UTCF."""
        dt = datetime(2001, 1, 1) + timedelta(seconds=met)
        if isutc and utcf is not None:
            dt += timedelta(seconds=utcf)
        ret = cls(dt.year, dt.month, dt.day, dt.hour, dt.minute, dt.second, dt.microsecond)
        ret.utcf = utcf
        ret.isutc = isutc
        return ret

    # Attribute aliases
    swift = swifttime
    utc = utctime



# The variable that indicates if a parameter is required
classReqPar = {
    "lc": "lc",
    "spec": "spec",
    "psf": "dopsf",
    "enh": "doenh",
    "xastrom": "doxastrom",
    "image": "image",
    "sourceDet": "sourceDet",
}


# Conversion from the short-form normally used to reference them, so a
# full name for printing
longProdName = {
    "lc": "light curve",
    "spec": "spectrum",
    "psf": "standard position",
    "enh": "enhanced position",
    "xastrom": "astrometric position",
    "image": "image",
    "sourceDet": "source detection",
}

# Conversion from the long name used by the Python interface, to my
# abbrevations
longToShort = {
    "LightCurve": "lc",
    "Spectrum": "spec",
    "StandardPos": "psf",
    "EnhancedPos": "enh",
    "AstromPos": "xastrom",
    "Image": "image",
    "SourceDet": "sourceDet",
}

# Conversion from my abbreviations to those used by the Python
# interface.
shortToLong = {
    "lc": "LightCurve",
    "spec": "Spectrum",
    "psf": "StandardPos",
    "enh": "EnhancedPos",
    "xastrom": "AstromPos",
    "image": "Image",
    "sourceDet": "SourceDet",
}

# downloadFormats lists supported download formats:
downloadFormats = ("zip", "tar", "tar.gz")



from .api_common import TOOAPI_Baseclass
from .api_daterange import TOOAPI_Daterange
from .api_resolve import TOOAPI_AutoResolve
from .api_skycoord import TOOAPI_SkyCoord
from .api_status import TOOStatus
from .swift_clock import TOOAPI_ClockCorrect


class Swift_VisWindow(TOOAPI_Baseclass, TOOAPI_ClockCorrect):
    """Simple class to define a Visibility window. Begin and End of window can
    either be accessed as self.begin or self.end, or as self[0] or self[1].

    Attributes
    ----------
    begin : datetime
        begin time of window
    end : datetime
        end time of window
    length : timedelta
        length of window
    """

    # API parameter definition
    _attributes = ["begin", "end", "length"]
    # Names for parameters
    _varnames = {"begin": "Begin Time", "end": "End Time", "length": "Window length"}
    # API name
    api_name = "Swift_VisWindow"

    def __init__(self):
        # Set all times in this class to UTC
        self._isutc = True
        # Attributes
        self.begin = None
        self.end = None

    @property
    def length(self):
        if self.end is None or self.begin is None:
            return None
        return self.end - self.begin

    @property
    def _table(self):
        header = [self._header_title(row) for row in self._parameters + self._attributes]
        return header, [[self.begin, self.end, self.length]]

    def __str__(self):
        return f"{self.begin} - {self.end} ({self.length})"

    def __getitem__(self, index):
        if index == 0:
            return self.begin
        elif index == 1:
            return self.end
        else:
            raise IndexError("list index out of range")


class Swift_VisQuery(
    TOOAPI_Baseclass,
    TOOAPI_Daterange,
    TOOAPI_SkyCoord,
    TOOAPI_AutoResolve,
    TOOAPI_ClockCorrect,
):
    """Request Swift Target visibility windows. These results are low-fidelity,
    so do not give orbit-to-orbit visibility, but instead long term windows
    indicates when a target is observable by Swift and not in a Sun/Moon/Pole
    constraint.

    Attributes
    ----------
    begin : datetime
        begin time of visibility window
    end : datetime
        end time of visibility window
    length : timedelta
        length of visibility window
    ra : float
        Right Ascension of target in J2000 (decimal degrees)
    dec : float
        Declination of target in J2000 (decimal degrees)
    skycoord : SkyCoord
        SkyCoord version of RA/Dec if astropy is installed
    hires : boolean
        Calculate visibility with high resolution, including Earth
        constraints
    username : str
        username for TOO API (default 'anonymous')
    shared_secret : str
        shared secret for TOO API (default 'anonymous')
    entries : list
        List of visibility windows (`Swift_VisWindow`)
    status : TOOStatus
        Status of API request
    """

    # Parameters that are passed to the API for this request
    _parameters = ["username", "ra", "dec", "length", "begin", "hires"]
    # Local and alias parameter names
    _local = ["name", "skycoord", "end", "shared_secret"]
    # Attributes returned by API Server
    _attributes = ["status", "windows"]
    # Subclasses
    _subclasses = [TOOStatus, Swift_VisWindow]
    # API Name
    api_name = "Swift_VisQuery"
    # Returned data
    windows = list()

    def __init__(self, *args, **kwargs):
        """
        Parameters
        ----------
        begin : datetime
            begin time of visibility window
        end : datetime
            end time of visibility window
        length : timedelta
            length of visibility window
        ra : float
            Right Ascension of target in J2000 (decimal degrees)
        dec : float
            Declination of target in J2000 (decimal degrees)
        skycoord : SkyCoord
            SkyCoord version of RA/Dec if astropy is installed
        hires : boolean
            Calculate visibility with high resolution, including Earth
            constraints
        username : str
            username for TOO API (default 'anonymous')
        shared_secret : str
            shared secret for TOO API (default 'anonymous')
        """
        # Set all times in this class to UTC
        self._isutc = True
        # User arguments
        self.username = "anonymous"
        self.ra = None
        self.dec = None
        self.hires = None
        self.length = None
        # Visibility windows go here
        self.windows = list()
        # Status of request
        self.status = TOOStatus()
        # Parse argument keywords
        self._parseargs(*args, **kwargs)

        if self.validate():
            self.submit()
        else:
            self.status.clear()

    @property
    def _table(self):
        if len(self.windows) != 0:
            header = self.windows[0]._table[0]
        else:
            header = []
        return header, [win._table[1][0] for win in self.windows]

    # For compatibility / consistency with other classes.
    @property
    def entries(self):
        return self.windows

    def __getitem__(self, index):
        return self.windows[index]

    def __len__(self):
        return len(self.windows)

    def validate(self):
        """Validate API submission before submit

        Returns
        -------
        bool
            Was validation successful?
        """
        # If length is not set, set to default of 7 days, or 1 day for hires
        if self.length is None:
            if self.hires:
                self.length = 1
            else:
                self.length = 7
        # Check RA/Dec are set correctly
        if self.ra is not None and self.dec is not None:
            if self.ra >= 0 and self.ra <= 360 and self.dec >= -90 and self.dec <= 90:
                return True
            else:
                self.status.error("RA/Dec not in valid range.")
                return False
        else:
            self.status.error("RA/Dec not set.")
            return False


# Shorthand alias for class
VisQuery = Swift_VisQuery
VisWindow = Swift_VisWindow
